# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
A module for .....
"""

from .analytics import (get_parent_modules_list, get_clone_modules_list, sensor_has_not_changed)  ## , run_clone_modules

__all__ = ['get_parent_modules_list', 'get_clone_modules_list', 'sensor_has_not_changed']  # 'run_clone_modules',



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import sys
import uuid
import logging
from ._authentication._authentication import get_switch_credentials
from ._utils._utils import ApiInputs, ApiHeaders
from ._utils._constants import SWITCH_ENVIRONMENT

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

consoleHandler = logging.StreamHandler(stream=sys.stdout)
# consoleHandler.setLevel(logging.INFO)

logger.addHandler(consoleHandler)
formatter = logging.Formatter('%(asctime)s  switch_api.%(module)s.%(funcName)s  %(levelname)s: %(message)s',
                              datefmt='%Y-%m-%dT%H:%M:%S')
consoleHandler.setFormatter(formatter)


def initialize(api_project_id: uuid.UUID, environment: SWITCH_ENVIRONMENT = 'Production') -> ApiInputs:
    """
    Initialize session with API when running scripts on your local machine.

    Parameters
    ----------
    api_project_id : uuid.UUID
        The ApiProjectID for the portfolio.
    environment: SWITCH_ENVIRONMENT
        (Optional) The Switch API environment to interact with

    Returns
    -------
    ApiInputs
        Returns the ApiInputs namedtuple that is required as an input to other functions.
    """

    try:
        uuid.UUID(api_project_id)
    except:
        sw.logger.error(f"Invalid value provided '{api_project_id}'- please provide a valid api_project_id. ")
        return False

    creds = get_switch_credentials(environment, api_project_id)

    logger.info("Successfully initialized.")
    api_base_url = f"https://{creds.api_endpoint}/api/1.0"
    # api_base_url = f"https://localhost:5001/api/1.0"
    api_projects_endpoint = f"{api_base_url}/projects"

    token = creds.api_token

    data_feed_id = '00000000-0000-0000-0000-000000000000'
    data_feed_file_status_id = '00000000-0000-0000-0000-000000000000'

    api_default = {
        'Content-Type': 'application/json; charset=utf-8',
        'Ocp-Apim-Subscription-Key': creds.subscription_key,
        'Authorization': f'bearer {token}'
    }

    operation_metadata = {
        'DataFeedId': data_feed_id,
        'DataFeedFileStatusId': data_feed_file_status_id
    }

    # API Header specific for Integration API endpoints
    api_integration = {
        'Content-Type': 'application/json; charset=utf-8',
        'Ocp-Apim-Subscription-Key': creds.subscription_key,
        'Authorization': f'bearer {token}',
        'X-Operation-Metadata': str(operation_metadata)
    }

    api_headers = ApiHeaders(default=api_default, integration=api_integration)

    return ApiInputs(email_address=creds.email, user_id=creds.user_id, api_project_id=api_project_id,
                     data_feed_id=data_feed_id, data_feed_file_status_id=data_feed_file_status_id, bearer_token=token,
                     api_base_url=api_base_url, api_projects_endpoint=api_projects_endpoint,
                     subscription_key=creds.subscription_key, api_headers=api_headers)



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
This module contains the helper functions for error handling.

The module contains three functions:
    - invalid_file_format() which should be used to validate the source file received against the expected schema and
      post any issues identified to the data feed dashboard.

    - post_errors() which is used to post errors (apart from those identified by the invalid_file_format() function) to
      the data feed dashboard.

    - validate_datetime() which checks whether the values of the datetime column(s) of the source file are valid. Any
      datetime errors identified by this function should be passed to the post_errors() function.

The validate_datetime() function can be used to validate the datetime column(s) of the source file. The output
`df_invalid_datetime` from this function should be passed to the post_errors() function. For example,

>>> import pandas as pd
>>> import switch_api as sw
>>> api_inputs = sw.initialize(api_project_id=api_project_id) # set api_project_id to the relevant portfolio
>>> test_df = pd.DataFrame({'DateTime':['2021-06-01 00:00:00', '2021-06-01 00:15:00', '', '2021-06-01 00:45:00'],
... 'Value':[10, 20, 30, 40], 'device_id':['xyz', 'xyz', 'xyz', 'xyz']})
>>> df_invalid_datetime, df = validate_datetime(df=test_df, datetime_col=['DateTime'], dt_fmt='%Y-%m-%d %H:%M:%S')
>>> if df_invalid_datetime.shape[0] != 0:
...     sw.error_handlers.post_errors(api_inputs, df_invalid_datetime, error_type='DateTime')

"""
import pandas
import pandera
import logging
import requests
import sys
import json
from typing import Optional, List, Union
from .._utils._constants import ERROR_TYPE, PROCESS_STATUS
from .._utils._utils import ApiInputs

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
consoleHandler = logging.StreamHandler(stream=sys.stdout)
consoleHandler.setLevel(logging.INFO)

logger.addHandler(consoleHandler)
formatter = logging.Formatter('%(asctime)s  switch_api.%(module)s.%(funcName)s  %(levelname)s: %(message)s',
                              datefmt='%Y-%m-%dT%H:%M:%S')
consoleHandler.setFormatter(formatter)


def invalid_file_format(api_inputs: ApiInputs, schema: pandera.DataFrameSchema, raw_df: pandas.DataFrame):
    """Validates the raw file format and posts any errors to data feed dashboard

    Parameters
    ----------
    api_inputs : ApiInputs
        Object returned by call to initialize()
    schema : pandera.DataFrameSchema
        The defined data frame schema object to be used for validation.
    raw_df : pandas.DataFrame
        The raw dataframe created by reading the file.

    Returns
    -------
    response_data_frame : pandas.DataFrame
        Data frame containing the response from API after posting any errors to data feed dashboard.

    """
    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    headers = api_inputs.api_headers.default

    error_type = 'FileFormat'
    process_status = 'Failed'

    try:
        schema.validate(raw_df, lazy=True)
    except pandera.errors.SchemaErrors as err:
        url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/data-feed/" \
              f"{api_inputs.data_feed_id}/errors?status={process_status}&errorType={error_type}&statusId=" \
              f"{api_inputs.data_feed_id}"
        logger.info("Sending request: POST %s", url)
        logger.error('Schema errors present: %s', err.failure_cases)

        schema_error = err.failure_cases

        response = requests.post(url, data=schema_error.to_json(orient='records'), headers=headers)
        response_status = '{} {}'.format(response.status_code, response.reason)
        if response.status_code != 200:
            logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                         response.reason)
            return response_status, pandas.DataFrame()
        elif len(response.text) == 0:
            logger.error('No data returned for this API call. %s', response.request.url)
            return response_status, pandas.DataFrame()

        response_data_frame = pandas.read_json(response.text)

        return response_data_frame


def post_errors(api_inputs: ApiInputs, errors: Union[pandas.DataFrame, str], error_type: ERROR_TYPE,
                process_status: PROCESS_STATUS = 'ActionRequired'):
    """Post errors to the Data Feed Dashboard

    Post dataframe containing the errors of type ``error_type`` to the Data Feed Dashboard in the Switch Platform.

    Parameters
    ----------
    api_inputs : ApiInputs
        The object returned by call to initialize() function.
    errors : Union[pandas.DataFrame, str]
        The dataframe containing the row(s) with errors or a string describing the error(s).
    error_type : ERROR_TYPE
        The type of error being posted to Data Feed Dashboard.
    process_status: PROCESS_STATUS, optional
        Set the status of the process to one of the allowed values specified by the PROCESS_STATUS literal
        (Default value = 'ActionRequired').

    Returns
    -------
    response_boolean: boolean
        True or False indicating the success of the call

    """
    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    if (api_inputs.data_feed_id == '00000000-0000-0000-0000-000000000000' or
            api_inputs.data_feed_file_status_id == '00000000-0000-0000-0000-000000000000'):
        logger.error("Post Errors can only be called in Production.")
        return False

    if not set([error_type]).issubset(set(ERROR_TYPE.__args__)):
        logger.error('error_type parameter must be set to one of the allowed values defined by the '
                     'ERROR_TYPE literal: %s', ERROR_TYPE.__args__)
        return False

    if not set([process_status]).issubset(set(PROCESS_STATUS.__args__)):
        logger.error('process_status parameter must be set to one of the allowed values defined by the '
                     'PROCESS_STATUS literal: %s', PROCESS_STATUS.__args__)
        return False

    headers = api_inputs.api_headers.default

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/data-feed/" \
          f"{api_inputs.data_feed_id}/errors?status={process_status}&errorType={error_type}&statusId=" \
          f"{api_inputs.data_feed_file_status_id}"
    logger.info("Sending request: POST %s", url)

    payload = ""
    if (isinstance(errors, pandas.DataFrame)):
        payload = json.dumps(json.loads(errors.to_json(orient='table'))['data'])
    elif (isinstance(errors, str)):
        payload = errors
    else:
        logger.error('errors_content parameter must be set to either a type of pandas.DataFrame or string')
        return False

    response = requests.post(url,
                             data=payload,
                             headers=headers)

    response_status = '{} {}'.format(response.status_code, response.reason)
    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status, False
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)
        return response_status, False

    logger.error(f'Directive:ProcessStatus={str(process_status)}')

    return response.text


def validate_datetime(df: pandas.DataFrame, datetime_col: Union[str, List[str]],
                      dt_fmt: Optional[str] = None):  # -> tuple[pandas.DataFrame, pandas.DataFrame]:
    """Check for datetime errors.

    Returns a tuple ``(df_invalid_datetime, df)``, where:
        - ``df_invalid_datetime`` is a dataframe containing the extracted rows of the input df with invalid datetime
        values.

        - ``df`` is the original dataframe input after dropping any rows with datetime errors.

    Parameters
    ----------
    df : pandas.DataFrame
        The dataframe that contains the columns to be validated.
    datetime_col: List[str]
        List of column names that contain the datetime values to be validated. If passing a single column name in as a
        string, it will be coerced to a list.
    dt_fmt : Optional[str]
        The expected format of the datetime columns to be coerced. The strftime to parse time, eg "%d/%m/%Y", note
        that "%f" will parse all the way up to nanoseconds. See strftime documentation for more information on
        choices: https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior

    Returns
    -------
    df_invalid_datetime, df : tuple[pandas.DataFrame, pandas.DataFrame]
        (df_invalid_datetime, df) - where: `df_invalid_date

    Notes
    -----
    If the ``df_invalid_datetime`` dataframe is not empty, then the dataframe should be passed to the ``post_errors()``
    function using ``error_type = 'DateTime'``. See example code below:

    >>> import pandas as pd
    >>> import switch_api as sw
    >>> api_inputs = sw.initialize(api_project_id = api_project_id) # set api_project_id to the relevant portfolio
    >>> test_df = pd.DataFrame({'DateTime':['2021-06-01 00:00:00', '2021-06-01 00:15:00', '', '2021-06-01 00:45:00'],
    ... 'Value':[10, 20, 30, 40], 'device_id':['xyz', 'xyz', 'xyz', 'xyz']})
    >>> df_invalid_datetime, df = validate_datetime(df=test_df, datetime_col=['DateTime'], dt_fmt='%Y-%m-%d %H:%M:%S')
    >>> if df_invalid_datetime.shape[0] != 0:
    ...     sw.error_handlers.post_errors(api_inputs, df_invalid_datetime, error_type='DateTime')

    """

    if type(datetime_col) == str:
        datetime_col = [datetime_col]
    elif type(datetime_col) == list:
        datetime_col = datetime_col
    else:
        logger.error('datetime_col: Invalid format - datetime_col must be a string or list of strings.')

    val_dt_cols = []
    for i in datetime_col:
        val_dt_cols.append(i + '_dt')
    lst = [None] * len(datetime_col)

    if dt_fmt is None:
        for i in range(len(datetime_col)):
            df[val_dt_cols[i]] = pandas.to_datetime((df[datetime_col[i]]), errors='coerce')
            lst[i] = df[df[val_dt_cols[i]].isnull()]
            df = df[df[val_dt_cols[i]].notnull()]
            lst[i] = lst[i].drop(val_dt_cols[i], axis=1)
            df = df.drop(val_dt_cols[i], axis=1)
        df_invalid_datetime = pandas.concat(lst, axis=0)
        df[datetime_col] = df[datetime_col].apply(lambda x: pandas.to_datetime(x))
        logger.info('Row count with invalid datetime values: %s', df_invalid_datetime.shape[0])
        logger.info('Row count with valid datetime values: %s', df.shape[0])
        return df_invalid_datetime, df
    else:
        for i in range(len(datetime_col)):
            df[val_dt_cols[i]] = pandas.to_datetime((df[datetime_col[i]]), errors='coerce', format=dt_fmt)
            lst[i] = df[df[val_dt_cols[i]].isnull()]
            df = df[df[val_dt_cols[i]].notnull()]
            lst[i] = lst[i].drop(val_dt_cols[i], axis=1)
            df = df.drop(val_dt_cols[i], axis=1)
        df_invalid_datetime = pandas.concat(lst, axis=0)
        df[datetime_col] = df[datetime_col].apply(lambda x: pandas.to_datetime(x, format=dt_fmt))
        logger.info('Row count with invalid datetime values: %s', df_invalid_datetime.shape[0])
        logger.info('Row count with valid datetime values: %s', df.shape[0])
        return df_invalid_datetime, df



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module defining the Automation class which contains register, deployment and helper methods.

----------
Automation
----------

This class contains the helper methods used to register, deploy, and test the created tasks. Additional helper functions
 for retrieving details of existing tasks on the Switch Automation platform are also included in this module.

"""
from typing import List, Union
import pandas
import requests
import inspect
import json
import logging
import sys
import uuid
from io import StringIO
from azure.servicebus import ServiceBusClient, ServiceBusReceiveMode  # , ServiceBusMessage
from .._utils._utils import _is_valid_regex, ApiInputs, _column_name_cap, DataFeedFileProcessOutput
from .._utils._constants import (argus_prefix, EXPECTED_DELIVERY, MAPPING_ENTITIES,
                                 QUEUE_NAME, ERROR_TYPE, SCHEDULE_TIMEZONE, DEPLOY_TYPE)
from .._utils._platform import _get_ingestion_service_bus_connection_string
from .pipeline import (Task, QueueTask, IntegrationTask, LogicModuleTask, AnalyticsTask, EventWorkOrderTask,
                       DiscoverableIntegrationTask)
from .definitions import (IntegrationDeviceDefinition, EventWorkOrderFieldDefinition, AnalyticsSettings,
                          IntegrationSettings, IntegrationDeviceConfigPropertyDefinition)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
consoleHandler = logging.StreamHandler(sys.stdout)
consoleHandler.setLevel(logging.INFO)

logger.addHandler(consoleHandler)
formatter = logging.Formatter('%(asctime)s  switch_api.%(module)s.%(funcName)s  %(levelname)s: %(message)s',
                              datefmt='%Y-%m-%dT%H:%M:%S')
consoleHandler.setFormatter(formatter)


class Automation:
    """Automation class defines the methods used to register and deploy tasks. """

    @staticmethod
    def run_queue_task(task: QueueTask, api_inputs: ApiInputs, consume_all_messages: bool = False):
        """Runs a Queue Task when in Development Mode

        The Queue Name should ideally be a testing Queue as messages will be consumed

        Parameters
        ----------
        task : QueueTask
            The custom QueueTask instance created by the user.
        api_inputs : ApiInputs
            Object returned by initialize() function.
        consume_all_messages : bool, default=False
            Consume all messages as they are read.

        Returns
        -------
        bool
            indicating whether the call was successful

        """

        if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
            logger.error("You must call initialize() before using API.")
            return pandas.DataFrame()

        if not issubclass(type(task), QueueTask):
            logger.error(
                "Driver must be an implementation of the QueueTask (Task).")
            return False

        logger.info("Running queue for  %s (%s) using queue name: %s", str(type(task).__name__), str(task.id),
                    str(task.queue_name))

        task.start(api_inputs)

        message_count = 0

        with ServiceBusClient.from_connection_string(
                _get_ingestion_service_bus_connection_string(api_inputs)) as client:
            with client.get_queue_receiver(task.queue_name,
                                           receive_mode=ServiceBusReceiveMode.RECEIVE_AND_DELETE) as receiver:
                while True:
                    messages = []
                    received_message_array = receiver.receive_messages(
                        max_message_count=task.maximum_message_count_per_call, max_wait_time=2)
                    for message in received_message_array:
                        messages.append(str(message))
                        message_count += 1

                    if len(received_message_array) == 0:
                        break

                    task.process_queue(api_inputs, messages)

                    if consume_all_messages == False:
                        break
        logger.info('Total messages consumed: %s', str(message_count))

    @staticmethod
    def reserve_instance(task: Task, api_inputs: ApiInputs, data_feed_id: uuid.UUID, minutes_to_reserve: int = 10):
        """Reserve a testing instance.

        Reserves a testing instance for the `driver`.

        Parameters
        ----------
        task : Task
            The custom Driver class created by the user.
        api_inputs : ApiInputs
            Object returned by initialize() function.
        data_feed_id : uuid.UUID
            The unique identifier of the data feed being tested.
        minutes_to_reserve : int, default = 10
                The duration in minutes that the testing instance will be reserved for (Default value = 10).

        Returns
        -------
        df : pandas.DataFrame
            Dataframe containing the details of the reserved testing instance.

        """

        if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
            logger.error("You must call initialize() before using API.")
            return pandas.DataFrame()

        headers = api_inputs.api_headers.default

        logger.info("Reserving a testing instance for %s (%s) on Data Feed Id (%s). ", str(type(task).__name__),
                    str(task.id), str(data_feed_id))

        url = argus_prefix + "ReserveInstance/" + str(data_feed_id) + "/" + str(minutes_to_reserve)
        logger.info("Sending request: GET %s", url)

        response = requests.request("GET", url, timeout=20, headers=headers)
        response_status = '{} {}'.format(response.status_code, response.reason)
        if response.status_code != 200:
            logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                         response.reason)
            return response_status, pandas.DataFrame()
        elif len(response.text) == 0:
            logger.error('No data returned for this API call. %s', response.request.url)
            return response_status, pandas.DataFrame()

        df = pandas.read_json(response.text, typ="Series")

        return df

    @staticmethod
    def register_task(api_inputs: ApiInputs, task: Union[Task, IntegrationTask, DiscoverableIntegrationTask, QueueTask,
                                                         AnalyticsTask, LogicModuleTask, EventWorkOrderTask]):
        """Register the task.

        Registers the task that was defined.

        Parameters
        ----------
        api_inputs : ApiInputs
            Object returned by initialize() function.
        task : Union[Task, IntegrationTask, DiscoverableIntegrationTask, QueueTask, AnalyticsTask, LogicModuleTask, EventWorkOrderTask]
            An instance of the custom class created from the Abstract Base Class `Task` or its abstract sub-classes:
            `IntegrationTask`,`DiscoverableIntegrationTask`, `AnalyticsTask`, `QueueTask`, `EventWorkOrderTask`, or
            `LogicModuleTask`.

        Returns
        -------
        pandas.DataFrame

        """

        if not issubclass(type(task), Task):
            logger.error(
                "Driver must be an implementation of the Abstract Base Class (Task).")
            return False

        base_class = ''
        allowed_base_classes = tuple([Task] + Task.__subclasses__())
        if issubclass(type(task), allowed_base_classes):
            base_class = task.__class__.__base__.__qualname__

        if base_class == '':
            logger.error(
                'Task must be an implementation of one of the Task sub-classes: IntegrationTask, LogicModuleTask, '
                'AnalyticsTask, EventWorkOrderTask, DiscoverableIntegrationTask.')
            return pandas.DataFrame()
        elif not set([base_class]).issubset(
                {allowed_base_classes[i].__qualname__ for i in range(len(allowed_base_classes))}):
            logger.error(f"Task must be an implementation of one of the Task sub-classes: IntegrationTask, "
                         f"LogicModuleTask, AnalyticsTask, EventWorkOrderTask, DiscoverableIntegrationTask.)")

        if type(task.mapping_entities) != list:
            logger.error('The mapping_entities parameter must have type = list. ')
            return pandas.DataFrame()
        elif not set(task.mapping_entities).issubset(set(MAPPING_ENTITIES.__args__)):
            logger.error('mapping_entities property must be a list containing one of the allowed values defined by the '
                         'MAPPING_ENTITIES literal: %s', MAPPING_ENTITIES.__args__)
            return pandas.DataFrame()

        task_settings = dict()

        if base_class == 'DiscoverableIntegrationTask':
            if not issubclass(type(task.integration_device_type_definition), IntegrationDeviceDefinition):
                logger.error(f"The integration_device_type_definition property must be a subclass of the "
                             f"IntegrationDeviceDefinition class. Current integration_device_type_definition property "
                             f"is a {type(task.integration_device_type_definition)}")
                return pandas.DataFrame()
            elif issubclass(type(task.integration_device_type_definition), IntegrationDeviceDefinition):
                props = task.integration_device_type_definition.config_properties

                if type(task.integration_device_type_definition.config_properties) != list:
                    logger.error(f"The task.integration_device_type_definition.config_properties must have type = list")
                    return pandas.DataFrame()
                else:
                    integration_settings_list = list()
                    if task.integration_device_type_definition.expose_address == True:
                        integration_settings_list.append({
                            'property_name': 'Address',
                            'display_name': task.integration_device_type_definition.address_label,
                            'editor': 'text_box', 'default_value': None, 'allowed_values': None})
                    prop_types = {True: [], False: []}
                    prop_issubclass = set()
                    for setting in task.integration_device_type_definition.config_properties:
                        val = issubclass(type(setting), IntegrationDeviceConfigPropertyDefinition)
                        prop_issubclass.add(val)
                        prop_types[val] += [setting]
                        integration_settings_list.append({
                            'property_name': setting.property_name, 'display_label': setting.display_label,
                            'editor': setting.editor, 'default_value': setting.default_value,
                            'allowed_values': setting.allowed_values})

                    if not prop_issubclass.issubset({True}):
                        logger.error(f"The task.integration_device_type_definition.config_properties parameter must be "
                                     f"a list containing one or more subclasses of the "
                                     f"IntegrationDeviceConfigPropertyDefinition class. {len(prop_types[False])} out of "
                                     f"the {len(task.integration_device_type_definition.config_properties)} items in the"
                                     f" list do not conform to this requirement. ")

            if len(integration_settings_list) > 0:
                task_settings['integration_settings'] = integration_settings_list

        if base_class == 'EventWorkOrderTask':
            # Validate that the work_order_fields_definition property matches required definition format.
            if type(task.work_order_fields_definition) != list:
                logger.error(f"The task.work_order_fields_definition property must have type list")
                return pandas.DataFrame
            else:
                prop_types = {True: [], False: []}
                prop_issubclass = set()
                work_order_input_list = list()
                for setting in task.work_order_fields_definition:
                    val = issubclass(type(setting), EventWorkOrderFieldDefinition)
                    prop_issubclass.add(val)
                    prop_types[val] += [setting]
                    work_order_input_list.append({'property_name': setting.property_name,
                                                  'display_label': setting.display_label,
                                                  'editor': setting.editor, 'default_value': setting.default_value,
                                                  'allowed_values': setting.allowed_values})
                if not prop_issubclass.issubset({True}):
                    logger.error(f"The task.work_order_fields_definition property must be a list containing one or "
                                 f"more subclasses of the EventWorkOrderFieldDefinition class. "
                                 f"{len(prop_types[False])} out of the {len(task.work_order_fields_definition)} items "
                                 f"in the list do not conform to this requirement. ")

            # Validate that the integration_settings_definition property matches required definition format.
            if type(task.integration_settings_definition) != list:
                logger.error(f"The task.integration_settings_definition property must have type list")
                return pandas.DataFrame
            else:
                prop_types = {True: [], False: []}
                prop_issubclass = set()
                integration_settings_list = list()
                for setting in task.integration_settings_definition:
                    val = issubclass(type(setting), IntegrationSettings)
                    prop_issubclass.add(val)
                    prop_types[val] += [setting]
                    integration_settings_list.append({'property_name': setting.property_name,
                                                      'display_label': setting.display_label,
                                                      'editor': setting.editor, 'default_value': setting.default_value,
                                                      'allowed_values': setting.allowed_values})
                if not prop_issubclass.issubset({True}):
                    logger.error(f"The task.integration_settings_definition property must be a list containing one or "
                                 f"more subclasses of the IntegrationSettings class. {len(prop_types[False])} "
                                 f"out of the {len(task.integration_settings_definition)} items in the list do not "
                                 f"conform to this requirement. ")
            if len(work_order_input_list) > 0 and len(integration_settings_list) > 0:
                task_settings['work_order_input'] = work_order_input_list
                task_settings['integration_settings'] = integration_settings_list

        if base_class == 'AnalyticsTask':
            # Validate that the analytics_settings_definition matches required definition format.
            if type(task.analytics_settings_definition) != list:
                logger.error(f"The task.analytics_settings_definition property must have type list")
                return pandas.DataFrame
            else:
                prop_types = {True: [], False: []}
                prop_issubclass = set()
                analytics_settings_list = list()
                for setting in task.analytics_settings_definition:
                    val = issubclass(type(setting), AnalyticsSettings)
                    prop_issubclass.add(val)
                    prop_types[val] += [setting]
                    analytics_settings_list.append({'property_name': setting.property_name,
                                                    'display_label': setting.display_label,
                                                    'editor': setting.editor, 'default_value': setting.default_value,
                                                    'allowed_values': setting.allowed_values})
                if not prop_issubclass.issubset({True}):
                    logger.error(f"The task.analytics_settings_definition property must be a list containing one or "
                                 f"more subclasses of the AnalyticsSettings class. "
                                 f"{len(prop_types[False])} out of the {len(task.analytics_settings_definition)} items "
                                 f"in the list do not conform to this requirement. ")

            if len(analytics_settings_list) > 0:
                task_settings['analytics_settings'] = analytics_settings_list

        logger.info("Registering %s (%s) to the Switch Driver Library: ", str(type(task).__name__), str(task.id))

        json_payload_verify = {
            "driverId": str(task.id),
            "name": type(task).__name__
        }

        if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
            logger.error("You must call initialize() before using API.")
            return pandas.DataFrame()

        headers = api_inputs.api_headers.default

        url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/drivers/verification"
        logger.info("Sending request: POST %s", url)

        response = requests.post(url, json=json_payload_verify, headers=headers)
        response_status = '{} {}'.format(response.status_code, response.reason)
        if response.status_code != 200:
            logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                         response.reason)
            return response_status, pandas.DataFrame()
        elif response.text == "Failed":
            logger.error("This Driver Guid is already associated with a Driver with a Different Name.")
            return response_status, pandas.DataFrame()

        json_payload = {
            "driverId": str(task.id),
            "name": type(task).__name__,
            "specification": task.description,
            "mappingEntities": task.mapping_entities,
            "scriptCode": Automation._get_task_code(task),
            "baseClass": base_class,
            "settings": task_settings
            # {"integration_settings":[{"property_name":"Dummy1", "display_name":"", "default_value":"", "allowed_values":[], "editor":""}], "work_order_input":[], "analytics_settings":[], }
        }

        url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/script-driver/registration"
        logger.info("Sending request: POST %s", url)

        response = requests.post(url, json=json_payload, headers=headers)
        response_status = '{} {}'.format(response.status_code, response.reason)
        if response.status_code != 200:
            logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                         response.reason)
            return response_status, pandas.DataFrame()
        elif len(response.text) == 0:
            logger.error('No data returned for this API call. %s', response.request.url)
            return response_status, pandas.DataFrame()

        df = pandas.read_json(response.text)
        df.columns = _column_name_cap(df.columns)

        return df

    @staticmethod
    def list_tasks(api_inputs: ApiInputs, search_name_pattern: str = '*'):
        """Get a list of the registered tasks.

        Parameters
        ----------
        api_inputs : ApiInputs
            Object returned by initialize() function.
        search_name_pattern : str, optional
            A pattern that should be used as a filter when retrieving the list of deployed drivers
            (Default value = '*').

        Returns
        -------
        pandas.DataFrame
            Dataframe containing the registered tasks.

        """
        if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
            logger.error("You must call initialize() before using API.")
            return pandas.DataFrame()

        headers = api_inputs.api_headers.default

        url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/drivers"

        if search_name_pattern != '*' and search_name_pattern != '':
            url += f"?name={search_name_pattern}"

        logger.info("Sending request: GET %s", url)

        response = requests.request("GET", url, timeout=20, headers=headers)
        response_status = '{} {}'.format(response.status_code, response.reason)
        if response.status_code != 200:
            logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                         response.reason)
            return response_status, pandas.DataFrame()
        elif len(response.text) == 0:
            logger.error('No data returned for this API call. %s', response.request.url)
            return response_status, pandas.DataFrame()

        df = pandas.read_json(response.text)
        df = df.drop(columns=['driverId'])
        df.columns = _column_name_cap(df.columns)

        return df

    @staticmethod
    def deploy_as_on_demand_data_feed(task: EventWorkOrderTask,
                                      api_inputs: ApiInputs, data_feed_id: uuid.UUID, settings: dict,
                                      queue_name: QUEUE_NAME = 'task', data_feed_name: str = None):
        """Deploy the custom driver as an on demand data feed.

        This deployment method is only suitable for the tasks that subclass the EventWorkOrderTask base class.

        Parameters
        ----------
        task : EventWorkOrderTask
            The custom driver class created from the Abstract Base Class `EventWorkOrderTask`
        api_inputs : ApiInputs
            Object returned by initialize() function.
        data_feed_id : uuid.UUID
            The DataFeedId to update if existing, else will create a new record with the given value.
        settings : dict
            List of settings used to deploy the driver. May containing information needed to access 3rd party api
             - e.g. URI, Username, Pwd, AccessToken, etc
        queue_name : QUEUE_NAME, optional
            The queue name (Default value = 'task').
        data_feed_name : str, Optional
            The name of the data feed (to be displayed in Task Insights UI). If not provided, will default to the
            task name.

        Returns
        -------
        df : pandas.DataFrame
            Dataframe containing the details of the deployed https endpoint data feed.

        """

        if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
            logger.error("You must call initialize() before using API.")
            return pandas.DataFrame()

        if not issubclass(task.__class__, EventWorkOrderTask):
            logger.error("Only task derived from EventWorkOrderTask can be deployed as OnDemand.")
            return pandas.DataFrame()

        if not isinstance(settings, dict):
            logger.error("setting parameter should be of type dict.")
            return pandas.DataFrame()
        elif not settings or not bool(settings) or len(settings) == 0:
            logger.error("setting parameter for On Demand Data Feed cannot be empty.")
            return pandas.DataFrame()

        if data_feed_id is None and data_feed_name is None:
            data_feed_id = uuid.UUID('00000000-0000-0000-0000-000000000000')
            data_feed_name = task.__class__.__qualname__
        elif data_feed_id is not None and data_feed_name is None:
            data_feed_id = data_feed_id
            data_feed_name = data_feed_name

        headers = api_inputs.api_headers.default

        if not set([queue_name]).issubset(set(QUEUE_NAME.__args__)):
            logger.error('queue_name parameter must be set to one of the allowed values defined by the '
                         'QUEUE_NAME literal: %s', QUEUE_NAME.__args__)
            return pandas.DataFrame()

        logger.info('Deploy %s (%s) as a data feed for ApiProjectID: %s', type(task).__name__, str(task.id),
                    api_inputs.api_project_id)

        payload = {
            "dataFeedId": str(data_feed_id),
            "driverId": str(task.id),
            "name": data_feed_name,
            "feedType": ",".join(task.mapping_entities),
            "expectedDelivery": "NA",
            "sourceType": "OnDemand",
            "queueName": queue_name,
            "settings": settings
        }

        url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/tasks/deployment"

        logger.info("Sending request: POST %s", url)

        response = requests.post(url, json=payload, headers=headers)
        response_status = '{} {}'.format(response.status_code, response.reason)
        if response.status_code != 200:
            logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                         response.reason)
            return response_status, pandas.DataFrame()
        elif len(response.text) == 0:
            logger.error('No data returned for this API call. %s', response.request.url)
            return response_status, pandas.DataFrame()

        df = pandas.read_json(StringIO(response.text), typ='Series').to_frame().T
        df.columns = _column_name_cap(df.columns)

        return df

    @staticmethod
    def deploy_as_email_data_feed(task: Union[Task, IntegrationTask, DiscoverableIntegrationTask, QueueTask,
                                              AnalyticsTask, LogicModuleTask, EventWorkOrderTask],
                                  api_inputs: ApiInputs, data_feed_id: uuid.UUID,
                                  expected_delivery: EXPECTED_DELIVERY, email_subject_regex: str,
                                  email_address_domain: str, queue_name: QUEUE_NAME = 'task',
                                  data_feed_name: str = None):
        """Deploy task as an email data feed.

        Deploys the created `task` as an email data feed. This allows the driver to ingest data sent via email. The
        data must be sent to data@switchautomation.com to be processed. If it is sent to another email address, the task
        will not be run.

        Parameters
        ----------
        task : Union[Task, IntegrationTask, DiscoverableIntegrationTask, QueueTask, AnalyticsTask, LogicModuleTask,
        EventWorkOrderTask]
            The custom driver class created from the Abstract Base Class `Task`
        api_inputs : ApiInputs
            Object returned by initialize() function.
        data_feed_id : uuid.UUID
            The DataFeedId to update if existing, else will create a new record with the given value.
        expected_delivery : EXPECTED_DELIVERY
            The expected delivery frequency.
        email_subject_regex : str
            Regex expression used to parse the email subject line to determine which driver the received file will
            be processed by.
        email_address_domain : str
            The email domain, without the @ symbol, of the sender. For example, if the email address that will send
            file(s) for this data feed to the Switch Automation Platform is sender@test.com, the string that should be
            passed to this parameter is "test.com".
        queue_name : QUEUE_NAME, Optional
            The name of queue (Default value = 'task').
        data_feed_name : str, Optional
            The name of the data feed (to be displayed in Task Insights UI). If not provided, the API will automatically
            default to using the task.name property of the `task` passed to function.


        Returns
        -------
        df : pandas.DataFrame
            Dataframe containing the details of the deployed email data feed.

        """

        if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
            logger.error("You must call initialize() before using API.")
            return pandas.DataFrame()

        if data_feed_id is None and data_feed_name is None:
            data_feed_id = uuid.UUID('00000000-0000-0000-0000-000000000000')
            data_feed_name = task.__class__.__qualname__
        elif data_feed_id is not None and data_feed_name is None:
            data_feed_id = data_feed_id
            data_feed_name = data_feed_name

        headers = api_inputs.api_headers.default

        logger.info('Deploy %s (%s) as a data feed for ApiProjectID: %s', type(task).__name__, str(task.id),
                    api_inputs.api_project_id)

        if not _is_valid_regex(email_subject_regex):
            logger.error("%s is not valid regex.", email_subject_regex)
            return pandas.DataFrame()

        if email_address_domain == "switchautomation.com":
            logger.info("Emails can only be received from the %s domain.", email_address_domain)

        if "@" in email_address_domain:
            logger.error("Do not include the @ in the email_address_domain parameter. ")
            return pandas.DataFrame()

        if not set([expected_delivery]).issubset(set(EXPECTED_DELIVERY.__args__)):
            logger.error('expected_delivery parameter must be set to one of the allowed values defined by the '
                         'EXPECTED_DELIVERY literal: %s', EXPECTED_DELIVERY.__args__)
            return pandas.DataFrame()

        if not set([queue_name]).issubset(set(QUEUE_NAME.__args__)):
            logger.error('queue_name parameter must be set to one of the allowed values defined by the '
                         'QUEUE_NAME literal: %s', QUEUE_NAME.__args__)
            return pandas.DataFrame()

        inbox_container = "data-exchange"
        payload = {
            "dataFeedId": str(data_feed_id),
            "driverId": str(task.id),
            "name": data_feed_name,
            "feedType": ",".join(task.mapping_entities),
            "expectedDelivery": expected_delivery,
            "sourceType": "Email",
            "queueName": queue_name,
            "email": {
                "emailAddressDomain": email_address_domain,
                "emailSubjectRegex": email_subject_regex,
                "container": inbox_container
            }
        }

        url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/tasks/deployment"

        logger.info("Sending request: POST %s", url)

        response = requests.post(url, json=payload, headers=headers)
        response_status = '{} {}'.format(response.status_code, response.reason)
        if response.status_code != 200:
            logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                         response.reason)
            return response_status, pandas.DataFrame()
        elif len(response.text) == 0:
            logger.error('No data returned for this API call. %s', response.request.url)
            return response_status, pandas.DataFrame()

        df = pandas.read_json(response.text, typ='Series').to_frame().T
        df.columns = _column_name_cap(df.columns)

        return df

    @staticmethod
    def deploy_as_ftp_data_feed(task: Union[Task, IntegrationTask, DiscoverableIntegrationTask, QueueTask,
                                            AnalyticsTask, LogicModuleTask, EventWorkOrderTask],
                                api_inputs: ApiInputs, data_feed_id: uuid.UUID, expected_delivery: EXPECTED_DELIVERY,
                                ftp_user_name: str, ftp_password: str, queue_name: QUEUE_NAME = 'task',
                                data_feed_name: str = None):
        """Deploy the custom driver as an FTP data feed

        Deploys the custom driver to receive data via an FTP data feed. Sets the `ftp_user_name` & `ftp_password` and
        the `expected_delivery` of the file.

        Parameters
        ----------
        task : Task
            The custom driver class created from the Abstract Base Class 'Task'
        api_inputs : ApiInputs
            Object returned by the initialize() function.
        data_feed_id : uuid.UUID
            The DataFeedId to update if existing, else will create a new record with the given value.
        expected_delivery : EXPECTED_DELIVERY
            The expected delivery frequency of the data.
        ftp_user_name : str
            The user_name to be used by the ftp service to authenticate delivery of the data feed.
        ftp_password : str
            The password to be used by the ftp service for the given `ftp_user_name` to authenticate delivery of the
            data feed.
        queue_name : QUEUE_NAME, default = 'task'
            The queue name (Default value = 'task').
        data_feed_name : str, Optional
            The name of the data feed (to be displayed in Task Insights UI). If not provided, will default to the
            task name.

        Returns
        -------
        df : pandas.DataFrame
            Dataframe containing the details of the deployed ftp data feed.

        """

        if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
            logger.error("You must call initialize() before using API.")
            return pandas.DataFrame()

        if data_feed_id is None and data_feed_name is None:
            data_feed_id = uuid.UUID('00000000-0000-0000-0000-000000000000')
            data_feed_name = task.__class__.__qualname__
        elif data_feed_id is not None and data_feed_name is None:
            data_feed_id = data_feed_id
            data_feed_name = data_feed_name

        headers = api_inputs.api_headers.default

        if not set([expected_delivery]).issubset(set(EXPECTED_DELIVERY.__args__)):
            logger.error('expected_delivery parameter must be set to one of the allowed values defined by the '
                         'EXPECTED_DELIVERY literal: %s', EXPECTED_DELIVERY.__args__)
            return pandas.DataFrame()

        if not set([queue_name]).issubset(set(QUEUE_NAME.__args__)):
            logger.error('queue_name parameter must be set to one of the allowed values defined by the '
                         'QUEUE_NAME literal: %s', QUEUE_NAME.__args__)
            return pandas.DataFrame()

        logger.info('Deploy %s (%s) as a data feed for ApiProjectID: %s', type(task).__name__, str(task.id),
                    api_inputs.api_project_id)

        inbox_container = "data-exchange"
        payload = {
            "dataFeedId": str(data_feed_id),
            "driverId": str(task.id),
            "name": data_feed_name,
            "feedType": ",".join(task.mapping_entities),
            "expectedDelivery": expected_delivery,
            "sourceType": "Ftp",
            "queueName": queue_name,
            "ftp": {
                "ftpUserName": ftp_user_name,
                "ftpPassword": ftp_password,
                "container": inbox_container
            }
        }

        url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/tasks/deployment"

        logger.info("Sending request: POST %s", url)

        response = requests.post(url, json=payload, headers=headers)
        response_status = '{} {}'.format(response.status_code, response.reason)
        if response.status_code != 200:
            logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                         response.reason)
            return response_status, pandas.DataFrame()
        elif len(response.text) == 0:
            logger.error('No data returned for this API call. %s', response.request.url)
            return response_status, pandas.DataFrame()

        df = pandas.read_json(response.text, typ='Series').to_frame().T
        df.columns = _column_name_cap(df.columns)

        return df

    @staticmethod
    def deploy_as_upload_data_feed(task: Union[Task, IntegrationTask, DiscoverableIntegrationTask, QueueTask,
                                               AnalyticsTask, LogicModuleTask, EventWorkOrderTask],
                                   api_inputs: ApiInputs, data_feed_id: uuid.UUID, expected_delivery: EXPECTED_DELIVERY,
                                   queue_name: QUEUE_NAME = 'task', data_feed_name: str = None):
        """Deploy the custom driver as a REST API end point Datafeed.

        To upload a file to the deployed data feed, use the UploadUrl from the response dataframe (with request type
        POST) with the following two headers:

        - 'Ocp-Apim-Subscription-Key' - set to the value of ``api_inputs.subscription_key``
        - 'Authorization' - set to the value 'Bearer ``api_inputs.bearer_token``'

        For example, to upload a file using the ``requests`` package:

        >>> import requests
        >>> url = df.loc[0,'UploadUrl']
        >>> payload={}
        >>> file_path = 'C:/xxyyzz.txt'
        >>> files={'file': open(file_path, 'rb')}
        >>> headers = {'Ocp-Apim-Subscription-Key': api_inputs.subscription_key, 'Authorization': f'Bearer {api_inputs.bearer_token}'}
        >>> response = requests.request("POST", url, headers=headers, data=payload, files=files)
        >>> print(response.text)

        Parameters
        ----------
        task : Task
            The custom driver class created from the Abstract Base Class 'Task'
        api_inputs : ApiInputs
            Object returned by the initialize() function.
        data_feed_id : uuid.UUID
            The DataFeedId to update if existing, else will create a new record with the given value.
        expected_delivery : EXPECTED_DELIVERY
            The expected delivery frequency of the data.
        queue_name : QUEUE_NAME, optional
            The queue name (Default value = 'task').
        data_feed_name : str, Optional
            The name of the data feed (to be displayed in Task Insights UI). If not provided, will default to the
            task name.

        Returns
        -------
        df : pandas.DataFrame
            Dataframe containing the details of the deployed https endpoint data feed.

        """

        if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
            logger.error("You must call initialize() before using API.")
            return pandas.DataFrame()

        if data_feed_id is None and data_feed_name is None:
            data_feed_id = uuid.UUID('00000000-0000-0000-0000-000000000000')
            data_feed_name = task.__class__.__qualname__
        elif data_feed_id is not None and data_feed_name is None:
            data_feed_id = data_feed_id
            data_feed_name = data_feed_name

        headers = api_inputs.api_headers.default

        if not set([expected_delivery]).issubset(set(EXPECTED_DELIVERY.__args__)):
            logger.error('expected_delivery parameter must be set to one of the allowed values defined by the '
                         'EXPECTED_DELIVERY literal: %s', EXPECTED_DELIVERY.__args__)
            return pandas.DataFrame()

        if not set([queue_name]).issubset(set(QUEUE_NAME.__args__)):
            logger.error('queue_name parameter must be set to one of the allowed values defined by the '
                         'QUEUE_NAME literal: %s', QUEUE_NAME.__args__)
            return pandas.DataFrame()

        logger.info('Deploy %s (%s) as a data feed for ApiProjectID: %s', type(task).__name__, str(task.id),
                    api_inputs.api_project_id)

        payload = {
            "dataFeedId": str(data_feed_id),
            "driverId": str(task.id),
            "name": data_feed_name,
            "feedType": ",".join(task.mapping_entities),
            "expectedDelivery": expected_delivery,
            "sourceType": "Upload",
            "queueName": queue_name,
            "upload": {
                "placeholder": ""
            },
        }

        url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/tasks/deployment"

        logger.info("Sending request: POST %s", url)

        response = requests.post(url, json=payload, headers=headers)
        response_status = '{} {}'.format(response.status_code, response.reason)
        if response.status_code != 200:
            logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                         response.reason)
            return response_status, pandas.DataFrame()
        elif len(response.text) == 0:
            logger.error('No data returned for this API call. %s', response.request.url)
            return response_status, pandas.DataFrame()

        df = pandas.read_json(StringIO(response.text), typ='Series').to_frame().T
        df.columns = _column_name_cap(df.columns)

        return df

    @staticmethod
    def deploy_on_timer(task: Union[Task, IntegrationTask, DiscoverableIntegrationTask, QueueTask,
                                    AnalyticsTask, LogicModuleTask, EventWorkOrderTask],
                        api_inputs: ApiInputs, data_feed_id: uuid.UUID, expected_delivery: EXPECTED_DELIVERY,
                        cron_schedule: str, queue_name: QUEUE_NAME = "task", settings: dict = None,
                        schedule_timezone: SCHEDULE_TIMEZONE = 'Local', timezone_offset_minutes: int = None,
                        data_feed_name: str = None):
        """Deploy driver to run on timer.

        Parameters
        ----------
        task : Task
            The custom driver class created from the Abstract Base Class `Task`.
        api_inputs : ApiInputs
            Object returned by initialize.initialize() function
        data_feed_id : uuid.UUID
            The DataFeedId to update if existing, else will create a new record with the given value.
        expected_delivery : EXPECTED_DELIVERY
            The expected delivery frequency.
        cron_schedule : str
            The CRONOS cron object containing the required schedule for the driver to be run. For details on the
            required format, see: https://crontab.cronhub.io/
        queue_name : QUEUE_NAME, optional
            The queue name (Default value = 'task').
        settings : dict, Optional
            List of settings used to deploy the driver. For example, may contain the user_name and password required to
            authenticate calls to a third-party API (Default value = None).
        schedule_timezone : SCHEDULE_TIMEZONE, optional
            Whether the ``cron_schedule`` should be applied based on Local or Utc timezone. If set to `Local`, this is
            taken as the timezone of the western-most site in the given portfolio (Default value = 'Local').
        timezone_offset_minutes: int, Optional
            Timezone offset in minutes (from UTC) to be used when applying the ``cron_schedule`` (Default value = None).
        data_feed_name : str, Optional
            The name of the data feed (to be displayed in Task Insights UI). If not provided, will default to the
            task name.

        Returns
        -------
        pandas.Dataframe
            A dataframe containing the details of the deployed data feed.

        """

        if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
            logger.error("You must call initialize() before using API.")
            return pandas.DataFrame()

        if data_feed_id is None and data_feed_name is None:
            data_feed_id = uuid.UUID('00000000-0000-0000-0000-000000000000')
            data_feed_name = task.__class__.__qualname__
        elif data_feed_id is not None and data_feed_name is None:
            data_feed_id = data_feed_id
            data_feed_name = data_feed_name

        if timezone_offset_minutes is None:
            timezone_offset_minutes = 0

        headers = api_inputs.api_headers.default

        if not set([expected_delivery]).issubset(set(EXPECTED_DELIVERY.__args__)):
            logger.error('expected_delivery parameter must be set to one of the allowed values defined by the '
                         'EXPECTED_DELIVERY literal: %s', EXPECTED_DELIVERY.__args__)
            return pandas.DataFrame()

        if not set([schedule_timezone]).issubset(set(SCHEDULE_TIMEZONE.__args__)):
            logger.error('schedule_timezone parameter must be set to one of the allowed values defined by the '
                         'SCHEDULE_TIMEZONE literal: %s', SCHEDULE_TIMEZONE.__args__)
            return pandas.DataFrame()

        if not set([queue_name]).issubset(set(QUEUE_NAME.__args__)):
            logger.error('queue_name parameter must be set to one of the allowed values defined by the '
                         'QUEUE_NAME literal: %s', QUEUE_NAME.__args__)
            return pandas.DataFrame()

        if 5 > len(cron_schedule.split(' ')) > 6:
            logger.error("cron_schedule parameter must be in the format * * * * *")
            return pandas.DataFrame()

        # if task.__class__.__qualname__ == 'DiscoverableIntegrationTask':
        #     config_props = task.integration_device_type_definition.config_properties
        #     property_name_dict = dict
        #     for i in range(len(config_props)):
        #         property_name_dict[config_props[i].property_name] = config_props[i].default_value

        # if task.__class__.__qualname__ == 'EventWorkOrderTask':
        #     config_props = task.integration_settings_definition
        #     property_name_list = []
        #     property_defaults_dict = {}
        #     for i in len(config_props):
        #         property_name_list.append(config_props[i].property_name)
        #         property_defaults_dict[config_props[i].property_name] = config_props[i].default_value
        #
        #     if settings is not None and not set(settings.keys()).issubset(set(property_name_list)):
        #         missing_keys = list(set(property_name_list).difference(set(settings.keys())))
        #         missing_required_keys = []
        #         missing_optional_keys = []
        #         for j in range(len(missing_keys)):
        #             if property_defaults_dict[missing_keys[j]] is None or property_defaults_dict[missing_keys[j]] == '':
        #                 missing_required_keys.append(missing_keys[j])
        #                 missing_optional_keys
        #             else:
        #                 missing_optional_keys.append(missing_keys[j])
        #                 missing_required_keys
        #         if len(missing_required_keys) > 0 and len(missing_optional_keys) == 0:
        #             loggger.error(f"settings parameter is missing the following required key(s): {missing_required_keys}.")

        logger.info('Deploy %s (%s) on timer for ApiProjectID: %s and schedule: %s.', type(task).__name__,
                    str(task.id), api_inputs.api_project_id, cron_schedule)
        logger.info('Feed Type is %s', str(task.mapping_entities))
        logger.info('Settings to be passed to the driver on start are: %s', str(settings))

        payload = {
            "dataFeedId": str(data_feed_id),
            "driverId": str(task.id),
            "name": data_feed_name,
            "feedType": ",".join(task.mapping_entities),
            "expectedDelivery": expected_delivery,
            "sourceType": "Timer",
            "queueName": queue_name,
            "timer": {
                "cronSchedule": cron_schedule,
                "timezoneOffsetMinutes": timezone_offset_minutes,
                "scheduleTimezone": schedule_timezone
            },
            "settings": settings
        }

        url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/tasks/deployment"

        logger.info("Sending request: POST %s", url)

        response = requests.post(url, json=payload, headers=headers)
        response_status = '{} {}'.format(response.status_code, response.reason)
        if response.status_code != 200 and len(response.text) > 0:
            logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                         response.reason)
            error_df = pandas.read_json(response.text)
            return response_status, error_df
        elif response.status_code != 200 and len(response.text) == 0:
            logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                         response.reason)
            return response_status, pandas.DataFrame()
        elif len(response.text) == 0:
            logger.error('No data returned for this API call. %s', response.request.url)
            return response_status, pandas.DataFrame()

        df = pandas.read_json(response.text, typ='Series').to_frame().T
        df.columns = _column_name_cap(df.columns)

        return df

    @staticmethod
    def cancel_deployed_data_feed(api_inputs: ApiInputs, data_feed_id: uuid.UUID, deployment_type: List[DEPLOY_TYPE]):
        """Cancel deployment for a given `data_feed_id` and `deployment_type`

        Parameters
        ----------
        api_inputs : ApiInputs
            Object returned by initialize.initialize() function
        data_feed_id: uuid.UUID
            Datafeed Id to cancel deployment
        deployment_type: List[DEPLOY_TYPE]


        Returns
        -------
        str
            A string containing the response text.

        """

        if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
            logger.error("You must call initialize() before using API.")
            return pandas.DataFrame()

        headers = api_inputs.api_headers.default

        if data_feed_id == None or data_feed_id == '':
            logger.error('Data feed Id cannot be Empty or None')
            return

        if type(deployment_type) != list:
            logger.error('deployment_type should be of type  list.')
            return
        elif deployment_type == None or len(deployment_type) == 0:
            logger.error('deployment_type cannot be empty. Please specify deployment_type to cancel deployment.')
            return
        elif not set(deployment_type).issubset(set(DEPLOY_TYPE.__args__)):
            logger.error('deployment_type item parameter must be set to one of the allowed values defined by the '
                         'DEPLOY_TYPE literal: %s', DEPLOY_TYPE.__args__)
            return

        url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/deployment/" \
              f"{data_feed_id}/cancel"

        logger.info("Cancel deployment type/s with DataFeedID: %s for ApiProjectID: %s", data_feed_id,
                    api_inputs.api_project_id)
        logger.info("Source Types: %s", ",".join(deployment_type))
        logger.info("Sending request: POST %s", url)

        payload = {
            'sourceTypes': deployment_type
        }

        response = requests.post(url, headers=headers, json=deployment_type)
        response_status = '{} {}'.format(response.status_code, response.reason)
        if response.status_code != 200:
            logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                         response.reason)
            return response_status, pandas.DataFrame()
        elif len(response.text) == 0:
            logger.error('No data returned for this API call. %s', response.request.url)
            return response_status, pandas.DataFrame()

        # Check if there's any timer deploy_type being cancelled, delete it as well by calling delete method
        if any("Timer" in deploy_type for deploy_type in deployment_type):
            return Automation.delete_deployed_data_feed(api_inputs=api_inputs, data_feed_id=data_feed_id)

        df = pandas.read_json(response.text, typ='Series').to_frame().T
        df.columns = _column_name_cap(df.columns)

        return df

    @staticmethod
    def delete_deployed_data_feed(api_inputs: ApiInputs, data_feed_id: uuid.UUID):
        """Delete deployment for a given `data_feed_id`.

        Parameters
        ----------
        api_inputs : ApiInputs
            Object returned by initialize.initialize() function
        data_feed_id: uuid.UUID
            Datafeed Id to cancel deployment

        Returns
        -------
        str
            A string containing the response text.

        """
        if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
            logger.error("You must call initialize() before using API.")
            return pandas.DataFrame()

        headers = api_inputs.api_headers.default

        if data_feed_id == None or data_feed_id == '':
            logger.error('Data feed Id cannot be None')
            return

        url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/deployment/" \
              f"{data_feed_id}/delete"

        logger.info("Delete deployment with DataFeedID: %s for ApiProjectID: %s", data_feed_id,
                    api_inputs.api_project_id)
        logger.info("Sending request: POST %s", url)

        response = requests.delete(url, headers=headers)
        response_status = '{} {}'.format(response.status_code, response.reason)
        if response.status_code != 200:
            logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                         response.reason)
            return response_status, pandas.DataFrame()
        elif len(response.text) == 0:
            logger.error('No data returned for this API call. %s', response.request.url)
            return response_status, pandas.DataFrame()

        df = pandas.read_json(response.text, typ='Series').to_frame().T
        df.columns = _column_name_cap(df.columns)

        return df

    @staticmethod
    def list_deployments(api_inputs: ApiInputs, search_name_pattern='*'):
        """Retrieve list of deployed drivers.

        Parameters
        ----------
        api_inputs : ApiInputs
            Object returned by initialize.initialize() function
        search_name_pattern : str
                A pattern that should be used as a filter when retrieving the list of deployed drivers
                (Default value = '*').

        Returns
        -------
        df : pandas.DataFrame
            Dataframe containing the drivers deployed for the given ApiProjectID that match the `search_name_pattern`.

        """

        if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
            logger.error("You must call initialize() before using API.")
            return pandas.DataFrame()

        headers = api_inputs.api_headers.default

        url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/data-feed/search"

        if search_name_pattern != '' and search_name_pattern != '*':
            url += f"?name={search_name_pattern}"

        logger.info("Sending request: GET %s", url)

        response = requests.request("GET", url, timeout=20, headers=headers)
        response_status = '{} {}'.format(response.status_code, response.reason)
        if response.status_code != 200:
            logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                         response.reason)
            return response_status, pandas.DataFrame()
        elif len(response.text) == 0:
            logger.error('No data returned for this API call. %s', response.request.url)
            return response_status, pandas.DataFrame()

        df = pandas.read_json(response.text)
        df.columns = _column_name_cap(df.columns)
        return df

    @staticmethod
    def list_data_feed_history(api_inputs: ApiInputs, data_feed_id: uuid.UUID, top_count: int = 10):
        """Retrieve data feed history

        Retrieves the `top_count` records for the given `data_feed_id`.

        Parameters
        ----------
        api_inputs : ApiInputs
            Object returned by initialize.initialize() function
        data_feed_id : uuid.UUID
            The unique identifier for the data feed that history should be retrieved for.
        top_count : int, default = 10
            The top record count to be retrieved. (Default value = 10).

        Returns
        -------
        df : pandas.DataFrame
            Dataframe containing the `top_count` history records for the given `data_feed_id`.

        """

        if data_feed_id is None or data_feed_id == '':
            logger.error('The data_feed_id parameter must be provided. ')
            return

        if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
            logger.error("You must call initialize() before using API.")
            return pandas.DataFrame()

        headers = api_inputs.api_headers.default

        url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/data-feed/" \
              f"{data_feed_id}/history?topCount={top_count}"

        logger.info("Sending request: GET %s", url)

        response = requests.request("GET", url, timeout=20, headers=headers)

        response_status = '{} {}'.format(response.status_code, response.reason)
        if response.status_code != 200:
            logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                         response.reason)
            return response_status, pandas.DataFrame()
        elif len(response.text) == 0:
            logger.error('No data returned for this API call. %s', response.request.url)
            return response_status, pandas.DataFrame()

        df = pandas.read_json(response.text)
        df.columns = _column_name_cap(df.columns)

        return df

    @staticmethod
    def data_feed_history_process_output(api_inputs: ApiInputs, data_feed_id: uuid.UUID = None,
                                         data_feed_file_status_id: uuid.UUID = None, row_number: int = None):
        """Retrieve data feed history process output

        Retrieves the `top_count` records for the given `data_feed_id`.

        Parameters
        ----------
        api_inputs : ApiInputs
            Object returned by initialize.initialize() function
        data_feed_id : uuid.UUID, default = None
            The unique identifier for the data feed that output should be retrieved for. This parameter works with the
            `row_number`
        data_feed_file_status_id : uuid.UUID, default = None
            The unique identifier for the data feed history should be retrieved for. This UUID can be retrieved from
            the list_data_feed_history() method.
        row_number: int, default = 0
            The row number from the list_data_feed_history method. It can be used in place of the file_status_id. Use
            row_number=0 to retrieve the most recent process.

        Returns
        -------
        log : str
            Containing the full print and output log for the process

        """

        if data_feed_id is None and data_feed_file_status_id is None:
            logger.error('Must supply either the data_feed_id + row_number, or the file_status_id')
            return
        elif data_feed_id is not None and data_feed_file_status_id is None and row_number is None:
            logger.error('When supplying the data_feed_id, you must also provide the row_number to retrieve. ')
            return
        elif data_feed_id is None and data_feed_file_status_id is not None and row_number is not None:
            logger.error('Please supply either the data_feed_id + row_number, or the file_status_id. ')
            return
        elif data_feed_id is not None and data_feed_file_status_id is not None and row_number is not None:
            logger.error('Please supply either the data_feed_id + row_number, or the file_status_id. ')
            return
        elif data_feed_id is not None and data_feed_file_status_id is not None and row_number is None:
            logger.error('Please supply either the data_feed_id + row_number or the file_status_id. ')
            return

        if row_number is None:
            row_number = -1

        if data_feed_id is None:
            data_feed_id = '00000000-0000-0000-0000-000000000000'

        if data_feed_file_status_id is None:
            data_feed_file_status_id = '00000000-0000-0000-0000-000000000000'

        if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
            logger.error("You must call initialize() before using API.")
            return pandas.DataFrame()

        headers = api_inputs.api_headers.default

        url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/data-feed/" \
              f"{data_feed_id}/file-status/{data_feed_file_status_id}/process-output?rowCount={row_number}"
        logger.info("Sending request: GET %s", url)

        response = requests.request("GET", url, timeout=20, headers=headers)

        response_status = '{} {}'.format(response.status_code, response.reason)
        if response.status_code != 200:
            logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                         response.reason)
            return response_status, pandas.DataFrame()
        elif len(response.text) == 0:
            logger.error('No data returned for this API call. %s', response.request.url)
            return response_status, pandas.DataFrame()

        data_feed_file_process_output = json.loads(response.text, object_hook=lambda d: DataFeedFileProcessOutput(
            data_feed_id=d['dataFeedId'], data_feed_file_status_id=d['fileStatusId'],
            client_tracking_id=d['clientTrackingId'],
            source_type=d['sourceType'], file_name=d['fileName'], file_received=d['fileReceived'],
            file_process_status=d['fileProcessStatus'], file_process_status_change=d['fileProcessStatusChange'],
            process_started=d['processStarted'], process_completed=d['processCompleted'],
            minutes_since_received=d['minutesSinceReceived'], minutes_since_processed=d['minutesSinceProcessed'],
            file_size=d['fileSize'], log_file_path=d['logFile'], output=d['output'], error=d['error']))

        return data_feed_file_process_output

    @staticmethod
    def data_feed_history_process_errors(api_inputs: ApiInputs, data_feed_id: uuid.UUID,
                                         data_feed_file_status_id: uuid.UUID):
        """Retrieve the unique error types for a given data feed file.

        Retrieves the distinct error types present for the given `data_feed_id` and `data_feed_file_status_id`.

        Parameters
        ----------
        api_inputs : ApiInputs
            Object returned by initialize.initialize() function
        data_feed_id : uuid.UUID
            The unique identifier for the data feed that errors should be retrieved for.
        data_feed_file_status_id : uuid.UUID
            The unique identifier for the file processed for a given data feed that errors should be retrieved for.

        Returns
        -------
        df : pandas.DataFrame
            Dataframe containing the distinct error types present for the given data_feed_file_status_id.

        """

        if data_feed_id == '' or data_feed_file_status_id == '':
            logger.error('The data_feed_id and data_feed_file_status_id parameters must be provided. ')
            return

        if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
            logger.error("You must call initialize() before using API.")
            return pandas.DataFrame()

        headers = api_inputs.api_headers.default

        url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/data-feed/" \
              f"{data_feed_id}/file-status/{data_feed_file_status_id}/process-errors"
        logger.info("Sending request: GET %s", url)

        response = requests.request("GET", url, timeout=20, headers=headers)

        response_status = '{} {}'.format(response.status_code, response.reason)
        if response.status_code != 200:
            logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                         response.reason)
            return response_status, pandas.DataFrame()
        elif len(response.text) == 0:
            logger.error('No data returned for this API call. %s', response.request.url)
            return response_status, pandas.DataFrame()

        df = pandas.read_json(response.text)
        df.columns = _column_name_cap(df.columns)

        return df

    @staticmethod
    def data_feed_history_errors_by_type(api_inputs: ApiInputs, data_feed_id: uuid.UUID,
                                         data_feed_file_status_id: uuid.UUID, error_type: ERROR_TYPE):
        """Retrieve the encountered errors for a given error type, data feed & file.

        Retrieves the errors identified for a given `error_type` for the `data_feed_id` and `data_feed_file_status_id`.

        Parameters
        ----------
        api_inputs : ApiInputs
            Object returned by initialize.initialize() function
        data_feed_id : uuid.UUID
            The unique identifier for the data feed that errors should be retrieved for.
        data_feed_file_status_id
            The unique identifier for the data feed that errors should be retrieved for.
        error_type: ERROR_TYPE
            The error type of the structured logs to be retrieved.

        Returns
        -------
        df : pandas.DataFrame
            Dataframe containing the errors identified for the given `data_feed_file_status_id`.

        """

        if (data_feed_id == '' or data_feed_file_status_id == '' or data_feed_id is None or
                data_feed_file_status_id is None):
            logger.error('The data_feed_id and data_feed_file_status_id parameters must be provided. ')
            return

        if not set([error_type]).issubset(set(ERROR_TYPE.__args__)):
            logger.error('error_type parameter must be set to one of the allowed values defined by the '
                         'ERROR_TYPE literal: %s', ERROR_TYPE.__args__)
            return pandas.DataFrame()

        headers = api_inputs.api_headers.default

        url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/data-feed/" \
              f"{data_feed_id}/file-status/{data_feed_file_status_id}/process-errors/type/{error_type}"

        logger.info("Sending request: GET %s", url)

        response = requests.request("GET", url, timeout=20, headers=headers)

        response_status = '{} {}'.format(response.status_code, response.reason)
        if response.status_code != 200:
            logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                         response.reason)
            return response_status, pandas.DataFrame()
        elif len(response.text) == 0:
            logger.error('No data returned for this API call. %s', response.request.url)
            return response_status, pandas.DataFrame()

        df = pandas.read_csv(StringIO(response.text), sep=",")  # , error_bad_lines=False)
        return df

    @staticmethod
    def _get_task_code(task):
        """

        Parameters
        ----------
        task : Task
            The custom driver class created from the Abstract Base Class `Task`.

        Returns
        -------
        driver_code : str
            The code used to create the `task`

        """
        task_type = type(task)
        task_code = ''
        parent_module = inspect.getmodule(task_type)
        for codeLine in inspect.getsource(parent_module).split('\n'):
            if codeLine.startswith('import ') or codeLine.startswith('from '):
                task_code += codeLine + '\n'

        task_code += '\n'
        task_code += inspect.getsource(task_type)
        task_code += ''
        task_code += 'task = ' + task_type.__name__ + '()'

        return task_code



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import json
import logging
import os
import sys
from typing import Optional, List

from msal import PublicClientApplication

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
consoleHandler = logging.StreamHandler(stream=sys.stdout)
consoleHandler.setLevel(logging.INFO)

logger.addHandler(consoleHandler)
formatter = logging.Formatter('%(asctime)s  %(name)s.%(funcName)s  %(levelname)s: %(message)s',
                              datefmt='%Y-%m-%dT%H:%M:%S')
consoleHandler.setFormatter(formatter)


def _merge_claims_challenge_and_capabilities(capabilities, claims_challenge):
    """
    Represent capabilities as {"access_token": {"xms_cc": {"values": capabilities}}} and then merge/add it into
    incoming claims
    """
    if not capabilities:
        return claims_challenge
    claims_dict = json.loads(claims_challenge) if claims_challenge else {}
    for key in ["access_token"]:  # We could add "id_token" if we'd decide to
        claims_dict.setdefault(key, {}).update(xms_cc={"values": capabilities})
    return json.dumps(claims_dict)


def _clean_up(result):
    if isinstance(result, dict):
        result.pop("refresh_in", None)  # MSAL handled refresh_in, customers need not
    return result


def _preferred_browser():
    """Register Edge and return a name suitable for subsequent webbrowser.get(...)
    when appropriate. Otherwise return None.
    """
    # On Linux, only Edge will provide device-based Conditional Access support
    if sys.platform != "linux":  # On other platforms, we have no browser preference
        return None
    browser_path = "/usr/bin/microsoft-edge"  # Use a full path owned by sys admin
    user_has_no_preference = "BROWSER" not in os.environ
    user_wont_mind_edge = "microsoft-edge" in os.environ.get("BROWSER", "")  # Note:
    # BROWSER could contain "microsoft-edge" or "/path/to/microsoft-edge".
    # Python documentation (https://docs.python.org/3/library/webbrowser.html)
    # does not document the name being implicitly register,
    # so there is no public API to know whether the ENV VAR browser would work.
    # Therefore, we would not bother examine the env var browser's type.
    # We would just register our own Edge instance.
    if (user_has_no_preference or user_wont_mind_edge) and os.path.exists(browser_path):
        try:
            import webbrowser  # Lazy import. Some distro may not have this.
            browser_name = "msal-edge"  # Avoid popular name "microsoft-edge"
            # otherwise `BROWSER="microsoft-edge"; webbrowser.get("microsoft-edge")`
            # would return a GenericBrowser instance which won't work.
            try:
                registration_available = isinstance(
                    webbrowser.get(browser_name), webbrowser.BackgroundBrowser)
            except webbrowser.Error:
                registration_available = False
            if not registration_available:
                logger.debug("Register %s with %s", browser_name, browser_path)
                # By registering our own browser instance with our own name,
                # rather than populating a process-wide BROWSER enn var,
                # this approach does not have side effect on non-MSAL code path.
                webbrowser.register(  # Even double-register happens to work fine
                    browser_name, None, webbrowser.BackgroundBrowser(browser_path))
            return browser_name
        except ImportError:
            pass  # We may still proceed
    return None


class CustomPublicClientApplication(PublicClientApplication):
    """
    Extends Msal PublicClientApplication to support custom Query Parameters.
    Future: Request feature in python msal GitHub
    """

    def __init__(self, client_id, client_credential=None, **kwargs):
        if client_credential is not None:
            raise ValueError("Public Client should not possess credentials")
        super(PublicClientApplication, self).__init__(
            client_id, client_credential=None, **kwargs)

    def acquire_token_interactive_custom(self, scopes: List[str], prompt=None, login_hint: str=None,
                                         domain_hint: str=None, claims_challenge: str=None, timeout: int=None,
                                         port: int=None, extra_scopes_to_consent=None, auth_params: dict=None, **kwargs):
        """Acquire token interactively i.e. via a local browser.
        Overridden to support custom query parameters.

        Prerequisite: In Azure Portal, configure the Redirect URI of your
        "Mobile and Desktop application" as ``http://localhost``.

        Parameters
        ----------
        scopes : List[str]
            It is a list of case-sensitive strings.
        prompt : str
            By default, no prompt value will be sent, not even "none". You will have to specify a value explicitly. Its
            valid values are defined in Open ID Connect specs
            https://openid.net/specs/openid-connect-core-1_0.html#AuthRequest
        login_hint : str
            Identifier of the user. Generally a User Principal Name (UPN) (Default value = None).
        domain_hint : str
            Can be one of "consumers" or "organizations" or your tenant domain "contoso.com". If included, it will skip
            the email-based discovery process that user goes through on the sign-in page, leading to a slightly more
            streamlined user experience. More information on possible values:
            `here <https://docs.microsoft.com/en-us/azure/active-directory/develop/v2-oauth2-auth-code-flow#request-an-
            authorization-code>`_ and
            `here <https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-oapx/
            86fb452d-e34a-494e-ac61-e526e263b6d8>`_.  (Default value = None)
        claims_challenge : str
            The claims_challenge parameter requests specific claims requested by the resource provider in the form of a
            claims_challenge directive in the www-authenticate header to be returned from the UserInfo Endpoint and/or
            in the ID Token and/or Access Token. It is a string of a JSON object which contains lists of claims being
            requested from these locations (Default value = None).
        timeout : int
            This method will block the current thread. This parameter specifies the timeout value in seconds. Default
            value ``None`` means wait indefinitely.
        port : int
            The port to be used to listen to an incoming auth response.
            By default we will use a system-allocated port.
            (The rest of the redirect_uri is hard coded as ``http://localhost``.)
        extra_scopes_to_consent : list
            "Extra scopes to consent" is a concept only available in AAD. It refers to other resources you might want
            to prompt to consent for, in the same interaction, but for which you won't get back a token for in this
            particular operation.
        auth_params : dict
            "Additional Auth Query Params to be sent to STS"

        Returns
        -------
        dict
            A dict containing no "error" key, and typically contains an "access_token" key. A dict containing an
            "error" key, when token refresh failed.

        """
        self._validate_ssh_cert_input_data(kwargs.get("data", {}))
        claims = _merge_claims_challenge_and_capabilities(
            self._client_capabilities, claims_challenge)
        telemetry_context = self._build_telemetry_context(
            self.ACQUIRE_TOKEN_INTERACTIVE)

        auth_params['claims'] = claims
        auth_params['domain_hint'] = domain_hint

        response = _clean_up(self.client.obtain_token_by_browser(
            scope=self._decorate_scope(scopes) if scopes else None,
            extra_scope_to_consent=extra_scopes_to_consent,
            redirect_uri="http://localhost:{port}".format(
                # Hardcode the host, for now. AAD portal rejects 127.0.0.1 anyway
                port=port or 0),
            prompt=prompt,
            login_hint=login_hint,
            timeout=timeout,
            auth_params=auth_params,
            data=dict(kwargs.pop("data", {}), claims=claims),
            headers=telemetry_context.generate_headers(),
            browser_name=_preferred_browser(),
            **kwargs))
        telemetry_context.update_telemetry(response)
        return response



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Module provides strategies to authenticate with the Switch Platform.
"""



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module defining the Task abstract base class which is inherited by the following specific task types:

- IntegrationTask
- DiscoverableIntegrationTask
- AnalyticsTask
- LogicModuleTask
- QueueTask
- EventWorkOrderTask

Also includes the Automation class which contains helper functions.

---------------
IntegrationTask
---------------
Base class used to create integrations between the Switch Automation platform and other platforms, low-level services,
or hardware.

Examples include:
    - Pulling readings or other types of data from REST APIs
    - Protocol Translators which ingest data sent to the platform via email, ftp or direct upload within platform.

---------------------------
DiscoverableIntegrationTask
---------------------------
Base class used to create integrations between the Switch Automation platform and 3rd party APIs.

Similar to the IntegrationTask, but includes a secondary method `run_discovery()` which triggers discovery of available
points on the 3rd party API and upserts these records to the Switch Platform backend so that the records are available
in Build - Discovery & Selection UI.

Examples include:
    - Pulling readings or other types of data from REST APIs

-------------
AnalyticsTask
-------------
Base class used to create specific analytics functionality which may leverage existing data from the platform. Each
task may add value to, or supplement, this data and write it back.

Examples include:
    - Anomaly Detection
    - Leaky Pipes
    - Peer Tracking

---------------
LogicModuleTask
---------------
Base class that handles the running, reprocessing and scheduling of the legacy logic modules in a way which enables
integration with other platform functionality.

---------
QueueTask
---------
Base class used to create data pipelines that are fed via a queue.

----------
Automation
----------
This class contains the helper methods used to register, deploy, and test the created tasks. Additional helper functions
 for retrieving details of existing tasks on the Switch Automation platform are also included in this module.

"""
import enum
import sys
# # import re
# import pandas
# import requests
# import inspect
import uuid
import logging
import datetime
# import json
from typing import List
from abc import ABC, abstractmethod
# from azure.servicebus import ServiceBusClient, ServiceBusReceiveMode  # , ServiceBusMessage
from .._utils._utils import ApiInputs, DiscoveryIntegrationInput  # , _is_valid_regex, generate_password, _column_name_cap, DataFeedFileProcessOutput
from .._utils._constants import (MAPPING_ENTITIES  # , api_prefix, argus_prefix, EXPECTED_DELIVERY, DEPLOY_TYPE,
                                 # QUEUE_NAME, ERROR_TYPE, SCHEDULE_TIMEZONE)
                                 )
from ..integration.helpers import get_templates, get_units_of_measure
from .definitions import (IntegrationDeviceDefinition, EventWorkOrderFieldDefinition, AnalyticsSettings, # BaseProperty
                          IntegrationSettings)  # , DeviceTypeDefinition
# from .._utils._platform import _get_ingestion_connection_string
# from io import StringIO

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
consoleHandler = logging.StreamHandler(sys.stdout)
consoleHandler.setLevel(logging.INFO)

logger.addHandler(consoleHandler)
formatter = logging.Formatter('%(asctime)s  switch_api.%(module)s.%(funcName)s  %(levelname)s: %(message)s',
                              datefmt='%Y-%m-%dT%H:%M:%S')
consoleHandler.setFormatter(formatter)


# Base Definitions in Switch package
class Task(ABC):
    """An Abstract Base Class called Task.

    Attributes
    ----------
    id : uuid.UUID
        Unique identifier of the task. This is an abstract property that needs to be overwritten when sub-classing.
        A new unique identifier can be created using uuid.uuid4()
    description : str
        Brief description of the task
    mapping_entities : List[MAPPING_ENTITIES]
        The type of entities being mapped. An example is:

        ``['Installations', 'Devices', 'Sensors']``
    author : str
        The author of the task.
    version : str
        The version of the task.

    """

    @property
    @abstractmethod
    def id(self) -> uuid.UUID:
        """Unique identifier of the task. Create a new unique identifier using uuid.uuid4() """
        pass

    @property
    @abstractmethod
    def description(self) -> str:
        """Brief description of the task"""
        pass

    @property
    @abstractmethod
    def mapping_entities(self) -> List[MAPPING_ENTITIES]:
        """The type of entities being mapped."""
        pass

    @property
    @abstractmethod
    def author(self) -> str:
        """"The author of the task."""
        pass

    @property
    @abstractmethod
    def version(self) -> str:
        """The version of the task"""
        pass


class IntegrationTask(Task):
    """Integration Task

    This class is used to create integrations that post data to the Switch Automation platform.


    Only one of the following methods should be implemented per class, based on the type of integration required:

    - process_file()
    - process_stream()
    - process()

    """

    @abstractmethod
    def process_file(self, api_inputs: ApiInputs, file_path_to_process: str):
        """Method to be implemented if a file will be processed by the Integration Task.

        The method should contain all code used to cleanse, reformat & post the data contained in the file.

        Parameters
        ----------
        api_inputs : ApiInputs
            object returned by call to initialize()
        file_path_to_process : str
            the file path

        """
        pass

    @abstractmethod
    def process_stream(self, api_inputs: ApiInputs, stream_to_process):
        """Method to be implemented if data received via stream

        The method should contain all code used to cleanse, reformat & post the data received via the stream.

        Parameters
        ----------
        api_inputs: ApiInputs
            The object returned by call to initialize()
        stream_to_process
            The details of the stream to be processed.
        """
        pass

    @abstractmethod
    def process(self, api_inputs: ApiInputs, integration_settings: dict):
        """Method to be implemented if data

        The method should contain all code used to cleanse, reformat & post the data pulled via the integration.

        Parameters
        ----------
        api_inputs: ApiInputs
            object returned by call to initialize()
        integration_settings : dict
            Any settings required to be passed to the integration to run. For example, username & password, api key,
            auth token, etc.

        """
        pass


class DiscoverableIntegrationTask(Task):
    """Discoverable Integration Task

    This class is used to create integrations that post data to the Switch Automation platform from 3rd party APIs that
    have discovery functionality.

    The `process()` method should contain the code required to post data for the integration. The `run_discovery()`
    method upserts records into the Integrations table so that end users can configure the points and import as
    devices/sensors within the Build - Discovery and Selection UI in the Switch Automation platform.

    Additional properties are required to be created to support both the discovery functionality & the subsequent
    device/sensor creation from the discovery records.

    """

    @property
    @abstractmethod
    def integration_device_type_definition(self) -> IntegrationDeviceDefinition:
        """The IntegrationDeviceDefinition used to create the DriverDevices records required for the Integration
        DeviceType to be available to drag & drop in the Build - Integration Schematic UI. Contains the properties that
        define the minimum set of required fields to be passed to the integration_settings dictionaries for the
        `process()` and `run_discovery()` methods"""
        pass

    # @property
    # @abstractmethod
    # def device_type_definitions(self) -> List[DeviceTypeDefinition]:
    #     """List of DeviceTypeDefinition classes used to create the Device Types available for selection in the
    #     Build - Discovery & Selection UI in the Switch Automation platform. """
    #     pass

    def check_integration_settings_valid(self, integration_settings: dict):
        required_integration_keys = set()

        if self.integration_device_type_definition.expose_address == True:
            required_integration_keys.add(self.integration_device_type_definition.address_label)

        for setting in self.integration_device_type_definition.config_properties:
            if setting.required_for_task == True:
                required_integration_keys.add(setting.property_name)

        if not required_integration_keys.issubset(set(integration_settings.keys())):
            logger.error(f'The integration_settings passed to the task do not contain the required keys: '
                         f'{required_integration_keys} ')
            return False
        else:
            return True

    @abstractmethod
    def run_discovery(self, api_inputs: ApiInputs, integration_settings: dict,
                      integration_input: DiscoveryIntegrationInput):
        """Method to implement discovery of available points from 3rd party API.

        The method should contain all code used to retrieve available points, reformat & post information to populate
        the Build - Discovery & Selection UI in the platform and allows users to configure discovered points prior to
        import.

        Parameters
        ----------
        api_inputs: ApiInputs
            object returned by call to initialize()
        integration_settings : dict
            Any settings required to be passed to the integration to run. For example, username & password, api key,
            auth token, etc.
        integration_input : DiscoveryIntegrationInput
            The information required to be sent to the container when the `run_discovery` method is triggered by a user
            from the UI. This information is the ApiProjectID, InstallationID, NetworkDeviceID and IntegrationDeviceID.

        """
        pass

    @abstractmethod
    def process(self, api_inputs: ApiInputs, integration_settings: dict):
        """Method to be implemented if data

        The method should contain all code used to cleanse, reformat & post the data pulled via the integration.

        Parameters
        ----------
        api_inputs: ApiInputs
            object returned by call to initialize()
        integration_settings : dict
            Any settings required to be passed to the integration to run. For example, username & password, api key,
            auth token, etc.

        """
        pass


class EventWorkOrderTask(Task):
    """Event Work Order Task

    This class is used to create work orders in 3rd party systems via tasks that are created in the Events UI of the
    Switch Automation platform.

    """

    @property
    @abstractmethod
    def work_order_fields_definition(self) -> List[EventWorkOrderFieldDefinition]:
        """Define the fields available in Events UI when creating a work order in 3rd Party System.

        The definition of the dictionary keys, display labels in Events UI, default value & allowed values
        for the generate_work_order() method's ``work_order_input`` parameter.

            property_name - the ``work_order_input`` dictionary key
            display_label - the display label for the given property_name in Events UI Work Order creation screen
            editor - the editor shown in Events UI Work Order creation screen
            default_value - default value for this property_name (if applicable)
            allowed_values - the set of allowed values (if applicable) for the given property_name. If editor=text_box,
            this should be None.

        """
        pass

    @property
    @abstractmethod
    def integration_settings_definition(self) -> List[IntegrationSettings]:
        """Define the generate_work_order() method's integration_settings dictionary requirements & defaults.

        The definition of the dictionary keys, display labels in Task Insights UI, default value & allowed values
        for the generate_work_order() method's ``integration_settings`` input parameter.

            property_name - the analytics_settings dictionary key
            display_label - the display label for the given property_name in Task Insights UI
            editor - the editor shown in Task Insights UI
            default_value - default value for this property_name (if applicable)
            allowed_values - the set of allowed values (if applicable) for the given property_name. If editor=text_box,
            this should be None.

        """
        pass

    def check_work_order_input_valid(self, work_order_input: dict):
        required_work_order_input_keys = set(['EventTaskId', 'Description', 'IntegrationId', 'DueDate'])

        for setting in self.work_order_fields_definition:
            required_work_order_input_keys.add(setting.property_name)

        if not required_work_order_input_keys.issubset(set(work_order_input.keys())):
            logger.error(f'The work_order_input passed to the task do not contain the required keys: '
                         f'{required_work_order_input_keys} ')
            return False
        else:
            return True

    def check_integration_settings_valid(self, integration_settings: dict):
        required_integration_keys = set()

        for setting in self.integration_settings_definition:
            required_integration_keys.add(setting.property_name)

        if not required_integration_keys.issubset(set(integration_settings.keys())):
            logger.error(f'The analytics_setting passed to the task do not contain the required keys: '
                         f'{required_integration_keys} ')
            return False
        else:
            return True

    @abstractmethod
    def generate_work_order(self, api_inputs: ApiInputs, integration_settings: dict, work_order_input: dict):
        """Generate work order in 3rd party system via Events UI

        Method to generate work order in 3rd party system based on a work order task created in Events UI in the Switch
        Automation platform.

        Notes
        -----
        In addition to the defined `work_order_fields_definition` fields, the `work_order_input` dictionary passed to
        this method will contain the following keys:

        - `EventTaskId`
            - unique identifier (uuid.UUID) for the given record in the Switch Automation platform
        - `Description`
            - free text field describing the work order to be generated.
        - `IntegrationId`
            - if linked to an existing work order in the 3rd party API, this will contain that system's identifier for
            the workorder. If generating a net new workorder, this field will be null (None).
        - `DueDate`
            - The due date for the work order as set in the Switch Automation platform.

        Parameters
        ----------
        api_inputs: ApiInputs
            object returned by call to initialize()
        integration_settings : dict
            Any settings required to be passed to the integration to run. For example, username & password, api key,
            auth token, etc.
        work_order_input : dict
            The work order defined by the task created in the Events UI of the Switch Automation platform. To be sent
            to 3rd party system for creation.
        """

        pass


class QueueTask(Task):
    """Queue Task

    This class is used to create integrations that post data to the Switch Automation platform using a Queue as the
    data source.

    Only one of the following methods should be implemented per class, based on the type of integration required:

    - process_queue()

    """

    @property
    @abstractmethod
    def queue_name(self) -> str:
        """The name of the queue to receive Data .. Name will actually be constructed as {ApiProjectId}_{queue_name} """
        pass

    @property
    @abstractmethod
    def maximum_message_count_per_call(self) -> int:
        """ The maximum amount of messages which should be passed to the process_queue at any one time
            set to zero to consume all
        """
        pass

    @abstractmethod
    def start(self, api_inputs: ApiInputs):
        """Method to be implemented if a file will be processed by the QueueTask Task.
        This will run once at the start of the processing and should contain

        """
        pass

    @abstractmethod
    def process_queue(self, api_inputs: ApiInputs, messages: List):
        """Method to be implemented if a file will be processed by the QueueTask Task.

        The method should contain all code used to consume the messages

        Parameter
        _________
        api_inputs : ApiInputs
            object returned by call to initialize()
        messages:List)
            list of serialized json strings which have been consumed from the queue

        """
        pass


class AnalyticsTask(Task):

    @property
    @abstractmethod
    def analytics_settings_definition(self) -> List[AnalyticsSettings]:
        """Define the start() method's analytics_settings dictionary requirements & defaults.

        The definition of the dictionary keys, display labels in Task Insights UI, default value & allowed values
        for the start() method's ``analytics_settings`` input parameter.

        property_name - the analytics_settings dictionary key
        display_label - the display label for the given property_name in Task Insights UI
        editor - the editor shown in Task Insights UI
        default_value - default value for this property_name (if applicable)
        allowed_values - the set of allowed values (if applicable) for the given property_name. If editor=text_box, this
        should be None.

        """
        pass

    def check_analytics_settings_valid(self, analytics_settings: dict):
        # required_analytics_settings_keys = set(['task_id'])
        required_analytics_settings_keys= set()

        for setting in self.analytics_settings_definition:
            required_analytics_settings_keys.add(setting.property_name)

        if not required_analytics_settings_keys.issubset(set(analytics_settings.keys())):
            logger.error(f'The analytics_setting passed to the task do not contain the required keys: '
                         f'{required_analytics_settings_keys} ')
            return False
        else:
            return True

    @abstractmethod
    def start(self, api_inputs: ApiInputs, analytics_settings: dict):
        """Start.

        The method should contain all code used by the task.

        Notes
        -----
        The method should first check the analytics_settings passed to the task are valid. Pseudo code below:
         >>> if self.check_analytics_settings_valid(analytics_settings=analytics_settings == True:
         >>>    # Your actual task code here - i.e. proceed with the task if valid analytics_settings passed.
         >>> else:
         >>>    sw.pipeline.logger.error('Invalid analytics_settings passed to driver. ')

        Parameters
        ----------
        api_inputs : ApiInputs
            the object returned by call to initialize()
        analytics_settings : dict
            any setting required by the AnalyticsTask

        """
        pass


class LogicModuleTask(Task):

    @abstractmethod
    def start(self, start_date_time: datetime.date, end_date_time: datetime.date, installation_id: uuid.UUID,
              share_tag_group: str, share_tags: list):
        pass

    # Needs to be implemented into the deploy_on_timer


class RunAt(enum.Enum):
    """ """
    Every15Minutes = 1
    Every1Hour = 3
    EveryDay = 4
    EveryWeek = 5





# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
A module for .....
"""
from azure.storage.blob import ContainerClient
import uuid
import pandas
import requests
import logging
import sys
from .._utils._constants import ACCOUNT
from .._utils._utils import ApiInputs

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
consoleHandler = logging.StreamHandler(stream=sys.stdout)
consoleHandler.setLevel(logging.INFO)

logger.addHandler(consoleHandler)
formatter = logging.Formatter('%(asctime)s  switch_api.%(module)s.%(funcName)s  %(levelname)s: %(message)s',
                              datefmt='%Y-%m-%dT%H:%M:%S')
consoleHandler.setFormatter(formatter)


class Queue:
    """ """
    @staticmethod
    def get_next_messages(account: ACCOUNT, container: str, queue_name: str, message_count: int = 1):
        """Retrieve next message(s).

        Parameters
        ----------
        account : ACCOUNT
            Azure account
        container : str
            Queue container.
        queue_name : str
            Queue name.
        message_count : int
            Message count to retrieve (Default value = 1).

        Returns
        -------

        """
        pass

    @staticmethod
    def send_message(account: ACCOUNT, container, queue_name, messages: list = None):
        """Send message

        Parameters
        ----------
        account : ACCOUNT
            Azure account.
        container : str
            Queue container.
        queue_name : str
            Queue name.
        messages : list, default = None
            Message (Default value = None).

        Returns
        -------

        """
        if messages is None:
            messages = []

        pass


class Blob:
    """ """
    @staticmethod
    def list(api_inputs: ApiInputs, account: ACCOUNT, container: str, prefix: str = None):
        """Retrieve list of blobs.

        Parameters
        ----------
        api_inputs : ApiInputs
            Object returned by initialize() function.
        account : ACCOUNT
            Azure account.
        container : str
            Blob container.
        prefix : str, default=None
            Prefix (Default value = None).

        Returns
        -------

        """

        if not set([account]).issubset(set(ACCOUNT.__args__)):
            logger.error('account parameter must be set to one of the allowed values defined by the '
                         'ACCOUNT literal: %s', ACCOUNT.__args__)
            return False

        if prefix is None:
            prefix = ''

        container_con_string = _get_container_sas_uri(api_inputs, container, account)
        container_client = ContainerClient.from_container_url(container_con_string)

        blob_list = container_client.list_blobs(name_starts_with=prefix)
        for blob in blob_list:
            logger.info('%s - %s', blob.name, str(blob.last_modified))
        return True

    @staticmethod
    def download(api_inputs: ApiInputs, account: ACCOUNT, container: str, blob_name: str):
        """

        Parameters
        ----------
        api_inputs : ApiInputs
            Object returned by initialize() function.
        account : ACCOUNT
            Azure account
        container : str
            Blob container.
        blob_name : str
            Blob name.

        Returns
        -------

        """
        if not set([account]).issubset(set(ACCOUNT.__args__)):
            logger.error('account parameter must be set to one of the allowed values defined by the '
                         'ACCOUNT literal: %s', ACCOUNT.__args__)
            return False

        container_con_string = _get_container_sas_uri(api_inputs, container, account)
        container_client = ContainerClient.from_container_url(container_con_string)

        blob_client = container_client.get_blob_client(blob_name)

        return blob_client.download_blob().readall()

    @staticmethod
    def upload(api_inputs: ApiInputs, data_frame: pandas.DataFrame, name: str, api_project_id, is_timeseries: bool,
               account: ACCOUNT = 'DataIngestion', batch_id: uuid.UUID = ''):
        """Upload data to blob.

        Parameters
        ----------
        api_inputs : ApiInputs
            Object returned by initialize() function.
        account: ACCOUNT, optional
            Blob account data will be uploaded to (Default value = 'SwitchStorage').
        data_frame : pandas.DataFrame
            Dataframe containing the data to be uploaded to blob.
        name : str
            Name.
        api_project_id : uuid.UUID
            ApiProjectID of the portfolio data is being uploaded for.
        is_timeseries : bool
            Define whether the data being uploaded is timeseries data or not.
        batch_id : str, default = ''
            Batch ID (Default value = '').

        Returns
        -------

        """
        chunk_size = 100000

        if not set([account]).issubset(set(ACCOUNT.__args__)):
            logger.error('account parameter must be set to one of the allowed values defined by the '
                         'ACCOUNT literal: %s', ACCOUNT.__args__)
            return False

        container = "data-ingestion-adx"
        container_con_string = _get_container_sas_uri(api_inputs, container, account, True)
        container_client = ContainerClient.from_container_url(container_con_string)

        list_chunked_df = [data_frame[count:count + chunk_size] for count in
                           range(0, data_frame.shape[0], chunk_size)]
        upload_path = "to-ingest/" + str(uuid.uuid1()) + "/" + name + "/"

        item_counter = 0
        for current_data_frame in list_chunked_df:
            item_counter += 1
            blob_name = upload_path + str(item_counter) + ".csv"
            logger.info("Uploading ... %s", blob_name)
            blob_client = container_client.get_blob_client(blob_name)
            data_csv = bytes(current_data_frame.to_csv(line_terminator='\r\n', index=False, header=False),
                             encoding='utf-8')
            blob_client.upload_blob(data_csv, blob_type="BlockBlob", overwrite=True)

        return upload_path, item_counter

    @staticmethod
    def custom_upload(api_inputs: ApiInputs, account: ACCOUNT, container: str, upload_path: str, file_name: str,
                      upload_object):
        """Upload data to blob.

        Parameters
        ----------
        api_inputs : ApiInputs
            Object returned by initialize() function.
        account: ACCOUNT
            Blob account data will be uploaded to.
        container : str
            Blob container.
        upload_path: str
            The prefix required to navigate from the base `container` to the folder the `upload_object` should be
            uploaded to.
        file_name : str
            File name to be stored in blob.
        upload_object :
            Object to be uploaded to blob.

        Returns
        -------

        """
        if not set([account]).issubset(set(ACCOUNT.__args__)):
            logger.error('account parameter must be set to one of the allowed values defined by the '
                         'ACCOUNT literal: %s', ACCOUNT.__args__)
            return False

        container_con_string = _get_container_sas_uri(api_inputs, container, account, True)
        container_client = ContainerClient.from_container_url(container_con_string)

        if upload_path.endswith('/') == False:
            blob_name = upload_path + '/' + file_name
        elif upload_path.endswith('/') == True:
            blob_name = upload_path + file_name

        logger.info('Uploading to blob: %s', blob_name)
        blob_client = container_client.get_blob_client(blob_name)
        blob_client.upload_blob(upload_object, blob_type="BlockBlob", overwrite=True)


def _get_ingestion_service_bus_connection_string(api_inputs: ApiInputs):
    """
    Get connection string specific to Data Ingestion Service Bus

    Parameters
    ----------
    api_inputs : ApiInputs
            Object returned by initialize() function.

    Returns
    -------
    str
        Data Ingestion Service Bus connection string
    """
    headers = api_inputs.api_headers.default

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/data-feed/service-bus"
    response = requests.request("GET", url, timeout=20, headers=headers)

    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return None
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)
        return None

    return response.text


def _get_container_sas_uri(api_inputs: ApiInputs, container: str, account: ACCOUNT = 'SwitcStorage', writable: bool = False):
    """
    Get container connection string from specified Storage Account

    Parameters
    ----------
    api_inputs : ApiInputs
            Object returned by initialize() function.
    container: str
        Name of the container under the account specified
    account : ACCOUNT, default = 'SwitchStorage'x
         (Default value = 'SwitchStorage')
    writable: bool
        Sets permissions expectation for the generated SAS Uri

    Returns
    -------
    str
        container connection string
    """

    if container == None or len(container) == 0:
        logger.error('Must set a container to get Container connection string.')

    headers = api_inputs.api_headers.default
    expire_after_hours = 1

    payload = {
        "storageOptions": __get_storage_options(account),
        "containerName": container,
        "expireAfterHours": expire_after_hours,
        "isWritable": writable
    }

    url = f"{api_inputs.api_base_url}/blob/container-sas"
    response = requests.request("POST", url, json=payload, headers=headers)

    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return ""
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)
        return ""

    return response.text


def _get_structure(df):
    """Get dataframe structure

    Parameters
    ----------
    df : pandas.DataFrame

    Returns
    -------

    """

    a = df.dtypes
    a = pandas.Series(a)
    a = a.reset_index().rename(columns={0: 'dtype'})
    a['dtype'] = a['dtype'].astype('str')
    a.set_index('index', inplace=True)
    a = a.to_dict()
    return a['dtype']


def __get_storage_options(account: ACCOUNT):
    """
    Get Storage Account Options. Currently the existing account literal doesn't match the Enum equivalent of storage
    options in the API, so we have this method

    Parameters
    ----------
    account : ACCOUNT
        Account to map API storage account options

    Returns
    -------
    str
        API equivalent account storage name
    """

    if account == None or len(account) == 0:
        logger.error('Mapping to storage options requires Account Parameter Value.')
        return account

    if account == 'SwitchStorage':
        return 'LegacySwitchStorage'
    elif account == 'SwitchContainer':
        return 'LegacySwitchContainer'
    elif account == 'Argus':
        return 'ArgusStorage'
    elif account == 'DataIngestion':
        return 'DataIngestionStorage'

    return account



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
A module for integrating asset creation, asset updates, data ingestion, etc into the Switch Automation platform.
"""

from .dataset import get_data, get_datasets_list, get_folders

__all__ = ['get_data', 'get_datasets_list', 'get_folders']



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
A module for .....
"""
# from io import StringIO
import pandas
import requests
import sys
import logging
import uuid
import datetime
from typing import List
# from .._utils._constants import api_prefix
from .._utils._utils import ApiInputs

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
consoleHandler = logging.StreamHandler(stream=sys.stdout)
consoleHandler.setLevel(logging.INFO)

logger.addHandler(consoleHandler)
formatter = logging.Formatter('%(asctime)s  switch_api.%(module)s.%(funcName)s  %(levelname)s: %(message)s',
                              datefmt='%Y-%m-%dT%H:%M:%S')
consoleHandler.setFormatter(formatter)


def get_parent_modules_list(api_inputs: ApiInputs, search_filter: str = ""):
    """Get list of data sets.

    Retrieves the list of parent modules either for the entire project across all installations
    or filter by name on the `search_filter`

    Parameters
    ----------
    api_inputs : ApiInputs
        Object returned by initialize() function.
    search_filter : str
        Uses contains on the parent module Name (Default value = "").

    Returns
    -------
    tuple (str, pandas.DataFrame)
        Returns the response status of the call and a dataframe containing the data returned by the call.

    """
    payload = {}

    headers = api_inputs.api_headers.default

    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/analytics/logic-modules"

    if search_filter != "*" and search_filter != '':
        url = url + f"?searchFilter={search_filter}"

    response = requests.request("GET", url, data=payload, timeout=20, headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)
    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status, pandas.DataFrame()
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)
        return response_status, pandas.DataFrame()

    df = pandas.read_json(response.text)
    logger.info("API Call successful.")

    return response_status, df


def get_clone_modules_list(api_inputs: ApiInputs, parent_module_id: uuid.UUID, search_filter="*"):
    """Get list of data sets.

    Retrieves the list of clone modules either for the specified parent_module_id
    or filter by shared Tag Name on the `search_filter`

    Parameters
    ----------
    api_inputs : ApiInputs
        Object returned by initialize() function.
    parent_module_id : uuid.UIID
        Parent logic module identifier
    search_filter : str
        Uses contains on the parent module Name (Default value = "*").

    Returns
    -------
    tuple (str, pandas.DataFrame)
        Returns the response status of the call and a dataframe containing the data returned by the call.

    """
    payload = {}

    headers = api_inputs.api_headers.default

    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/analytics/logic-modules-clones/" \
          f"parent-module/{parent_module_id}"

    if search_filter != "*" and search_filter != '':
        url = url + f"?searchFilter={search_filter}"

    response = requests.request("GET", url, data=payload, timeout=20, headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)
    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status, pandas.DataFrame()
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)
        return response_status, pandas.DataFrame()

    df = pandas.read_json(response.text)
    logger.info("API Call successful.")

    return response_status, df


# def run_clone_modules(api_inputs: ApiInputs, parent_module_id: uuid.UUID, start_date: datetime.date,
#                       end_date: datetime.date, share_to_tags: List[str] = None):
#     """Get list of data sets.
#
#     Run the list of clone modules for the specified parent_module_id
#     and the share to tags list
#
#     Parameters
#     ----------
#     api_inputs : ApiInputs
#         Object returned by initialize() function.
#     parent_module_id : uuid.UUID
#         Identifier for parent module.
#     start_date : datetime.date
#         Date the reprocess should start at.
#     end_date : datetime.date
#         Date the reprocess should finish at.
#     share_to_tags : List[str], optional
#         List of Tag names received via the get_clone_modules_list method
#
#     Returns
#     -------
#     tuple (str, pandas.DataFrame)
#         Returns the response status of the call and a dataframe containing the data returned by the call.
#
#     """
#     if start_date > end_date:
#         logger.error('start_date must be prior to end_date: start_date= %s, end_date= %s', start_date.__str__(),
#                      end_date.__str__())
#         return pandas.DataFrame()
#
#     payload = {
#         "parentModuleId": parent_module_id,
#         "dateFrom": start_date.isoformat(),
#         "dateTo": end_date.isoformat(),
#         "tagsList": share_to_tags
#     }
#
#     headers = {
#         'x-functions-key': api_inputs.api_key,
#         'Content-Type': 'application/json; charset=utf-8',
#         'user-key': api_inputs.user_id
#     }
#
#     if api_inputs.datacentre == '' or api_inputs.api_key == '':
#         logger.error("You must call initialize() before using API.")
#         return pandas.DataFrame()
#
#     url = api_prefix + api_inputs.datacentre + "/" + api_inputs.api_project_id + "/RunLogicModule"
#     response = requests.post(url, json=payload, headers=headers)
#     response_status = '{} {}'.format(response.status_code, response.reason)
#     if response.status_code != 200:
#         logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
#                      response.reason)
#         return response_status, pandas.DataFrame()
#     elif len(response.text) == 0:
#         logger.error('No data returned for this API call. %s', response.request.url)
#         return response_status, pandas.DataFrame()
#
#     table_data = StringIO(response.text)
#     df = pandas.read_table(table_data, sep=',')
#     logger.info("API Call successful.")
#     return response_status, df


def sensor_has_not_changed(api_inputs: ApiInputs, start_date: datetime.date, end_date: datetime.date,
                           templates: List[str] = None, hours_unchanged: int = 24):
    """Get list of sensors which haven't changed value for the given period and at least ``hours_unchanged``.

    Run the Has Not Changed Query for Selected List of Sites (empty List means All)
    and Selected List of Templates  (empty List means All)

    Parameters
    ----------
    api_inputs : ApiInputs
        Object returned by initialize() function.
    start_date : datetime.date
        Date the reprocess should start at.
    end_date : datetime.date
        Date the reprocess should finish at.
    templates : List[str]
        List of ObjectPropertyTemplateNames
    hours_unchanged: int
        number of hours the sensor value has not changed

    Returns
    -------
    tuple (str, pandas.DataFrame)
        Returns the response status of the call and a dataframe containing the data returned by the call as a tuple.

    """

    headers = api_inputs.api_headers.default

    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    if templates.count == 0:
        logger.error("You must supply templates to report on.")
        return pandas.DataFrame()

    payload = {
        "dateFrom": start_date.isoformat(),
        "dateTo": end_date.isoformat(),
        "templatesList": templates,
        "hoursUnchanged": hours_unchanged
    }

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/analytics/run-adx-query/SensorHasNotChanged"

    response = requests.post(url, json=payload, headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)
    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status, pandas.DataFrame()
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)
        return response_status, pandas.DataFrame()

    df = pandas.read_json(response.text)
    logger.info("API Call successful.")
    return response_status, df



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
A module for .....
"""
# import datetime
import re
import uuid
from collections import namedtuple
from typing import Union, List
import string
import secrets
import pandas
import pandera
import logging
import sys
from .._utils._constants import WORK_ORDER_STATUS, WORK_ORDER_CATEGORY, WORK_ORDER_PRIORITY

__all__ = ['generate_password', 'convert_to_sqm', 'convert_to_pascal_case', 'ValidatedSetterProperty', 'ApiInputs',
           'DiscoveryIntegrationInput', 'DataFeedFileProcessOutput']

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
consoleHandler = logging.StreamHandler(stream=sys.stdout)
consoleHandler.setLevel(logging.INFO)

logger.addHandler(consoleHandler)
formatter = logging.Formatter('%(asctime)s  switch_api.%(module)s.%(funcName)s  %(levelname)s: %(message)s',
                              datefmt='%Y-%m-%dT%H:%M:%S')
consoleHandler.setFormatter(formatter)


class ValidatedSetterProperty:
    def __init__(self, func, name=None, doc=None):
        self.func = func
        self.__name__ = name if name is not None else func.__name__
        self.__doc__ = doc if doc is not None else func.__doc__

    def __set__(self, obj, value):
        ret = self.func(obj, value)
        obj.__dict__[self.__name__] = value


def _column_name_cap(columns) -> list:
    renamed_columns = [name[0].upper() + name[1:] for name in columns.to_list()]
    return renamed_columns


def _with_func_attrs(**attrs):
    def with_attrs(f):
        for k, v in attrs.items():
            setattr(f, k, v)
        return f

    return with_attrs


def _is_valid_regex(regex):
    """Check if regex is valid.

    Parameters
    ----------
    regex : str
        A `regex` expression to be validated.

    Returns
    -------
    bool
        Valid if True, invalid if False.

    """
    try:
        re.compile(regex)
        return True
    except re.error:
        return False


ApiHeaders = namedtuple('ApiHeaders', ['default', 'integration'])

ApiInputs = namedtuple('ApiInputs',
                       ['email_address', 'user_id', 'api_project_id', 'data_feed_id', 'data_feed_file_status_id',
                        'bearer_token', 'api_base_url', 'api_projects_endpoint', 'subscription_key', 'api_headers'])

DiscoveryIntegrationInput = namedtuple(
        'DiscoveryIntegrationInput',
        ['api_project_id', 'installation_id', 'network_device_id', 'integration_device_id', 'user_id', 'batch_no'])
# DiscoveryIntegrationInput.__doc__ = """DiscoveryIntegrationInput(api_project_id, installation_id, network_device_id, integration_device_id, user_id, batch_no)
#
# Defines the required inputs to be provided to the run_discovery() method's integration_input parameter.
#
# Parameters
# ----------
# api_project_id : the unique identifier for the portfolio (ApiProjectID)
# installation_id : the unique identifier for the given site (InstallationID)
# network_device_id : the DeviceSpecs.NetworkDeviceID for the given integration_device_id
# integration_device_id : the unique identifier for the integration device the discovery is triggered against
# user_id : the unique identifier of the user who triggered the discovery
# batch_no : the unique identifier for the instance of the triggered discovery
# """
# DiscoveryIntegrationInput.api_project_id.__doc__ = "uuid.UUID : the unique identifier for the portfolio (ApiProjectID)"
# DiscoveryIntegrationInput.installation_id.__doc__ = "uuid.UUID : the unique identifier for the given site (InstallationID)"
# DiscoveryIntegrationInput.network_device_id.__doc__ = "uuid.UUID : the DeviceSpecs.NetworkDeviceID for the given integration_device_id"
# DiscoveryIntegrationInput.integration_device_id.__doc__ = "uuid.UUID : the unique identifier for the integration device the discovery is triggered against"
# DiscoveryIntegrationInput.user_id.__doc__ = "uuid.UUID : the unique identifier of the user who triggered the discovery"
# DiscoveryIntegrationInput.batch_no.__doc__ = "uuid.UUID : the unique identifier for the instance of the triggered discovery"

DataFeedFileProcessOutput = namedtuple('DataFeedFileProcessOutput',
                                       [
                                           'data_feed_id', 'data_feed_file_status_id',
                                           'client_tracking_id',
                                           'source_type', 'file_name', 'file_received',
                                           'file_process_status',
                                           'file_process_status_change', 'process_started',
                                           'process_completed',
                                           'minutes_since_received', 'minutes_since_processed',
                                           'file_size',
                                           'log_file_path', 'output', 'error'
                                       ])


def generate_password():
    """Generate a ten-character alphanumeric password with at least one lowercase character, at least one uppercase
    character, and at least three digits.
    """
    alphabet = string.ascii_letters + string.digits
    while True:
        password = ''.join(secrets.choice(alphabet) for i in range(10))
        if (any(c.islower() for c in password)
                and any(c.isupper() for c in password)
                and sum(c.isdigit() for c in password) >= 3):
            break
    return password


def convert_to_sqm(df: pandas.DataFrame, sqft_col_name: str):
    """Convert to square metres

    Convert floor area from square feet to square metres.

    Parameters
    ----------
    df : pandas.DataFrame
        The site asset register dataframe.
    sqft_col_name : str
        The name of the column containing the Floor Area values in sq. ft. to be converted.

    Returns
    -------
    df : pandas.DataFrame
        The input dataframe after converting the `sqft_col_name` column to square metres.


    """
    df[sqft_col_name] = df[sqft_col_name] * 0.09290304
    return df


def convert_to_pascal_case(text: Union[str, List[str]]):
    """Convert to PascalCase

    Capitalises the first letter of each word and strips any non-alphanumeric (``[a-zA-Z0-9]``) characters from between
    words as well as any leading or trailing the string.

    Parameters
    ----------
    text : Union[str, List[str]]
        A string or list of strings

    Returns
    -------
    Union[str, List[str]]
        If passed a string, the function returns the converted string. If passed a list of strings, the function
        returns a list of converted strings.

    """
    if type(text) == str:
        text = re.sub(r"^[^a-zA-Z0-9]", '', str(text))
        text = str.upper(text[0]) + text[1:]
        if not text:
            return text
        text = re.sub(r"[^a-zA-Z0-9]([a-zA-Z0-9])", lambda matched: str.upper(matched.group(1)), text)
        return re.sub(r"[^a-zA-Z0-9]", '', text)
    elif type(text) == list and {isinstance(text[j], str) for j in range(len(text))}.issubset([True]):
        for i in range(len(text)):
            text_item = text[i]

            text_item = re.sub(r"^[^a-zA-Z0-9]", '', str(text_item))
            text_item = str.upper(text_item[0]) + text_item[1:]
            if not text_item:
                text[i] = text_item
            text_item = re.sub(r"[^a-zA-Z0-9]([a-zA-Z0-9])", lambda matched: str.upper(matched.group(1)), text_item)

            text[i] = re.sub(r"[^a-zA-Z0-9]", '', text_item)
        return text
    else:
        logger.exception(f"The value passed to the text parameter was not a string or a list of strings. "
                         f"Unable to convert to pascal case. ")
        return False


_site_schema = pandera.DataFrameSchema(
    {
        'InstallationName': pandera.Column(pandera.String, checks=pandera.Check.str_length(1, 255)),
        'InstallationCode': pandera.Column(pandera.STRING, coerce=True, checks=pandera.Check.str_length(1, 100)),
        'Address': pandera.Column(pandera.String, checks=[pandera.Check.str_length(1, 250)]),
        'Country': pandera.Column(pandera.String, checks=[pandera.Check.str_length(1, 50)]),
        'Suburb': pandera.Column(pandera.String, checks=[pandera.Check.str_length(1, 50)]),
        'State': pandera.Column(pandera.String, checks=[pandera.Check.str_length(1, 100)], required=False),
        'StateName': pandera.Column(pandera.String, checks=[pandera.Check.str_length(1, 50, n_failure_cases=None)]),
        'FloorAreaM2': pandera.Column(pandera.Float, checks=[pandera.Check.greater_than_or_equal_to(0)]),
        'ZipPostCode': pandera.Column(pandera.STRING, checks=[pandera.Check.str_length(1, 20)]),
        'Latitude': pandera.Column(pandera.Float, required=False),
        'Longitude': pandera.Column(pandera.Float, required=False),
        'Timezone': pandera.Column(pandera.String, required=False),
        'InstallationId': pandera.Column(pandera.String, required=False),
    }
)

_work_order_schema = pandera.DataFrameSchema(
    columns={
        'WorkOrderId': pandera.Column(pandera.String, required=True),
        'InstallationId': pandera.Column(pandera.String, checks=[
            pandera.Check(lambda x: uuid.UUID(x).__class__ == uuid.UUID, element_wise=True)], required=True),
        'WorkOrderSiteIdentifier': pandera.Column(pandera.STRING, coerce=True, required=True),
        'WorkOrderCategory': pandera.Column(pandera.String, checks=[
            pandera.Check.isin(list(WORK_ORDER_CATEGORY.__args__), ignore_na=False)], required=True),
        'Type': pandera.Column(pandera.String, required=True),
        'Description': pandera.Column(pandera.String, required=True),
        'Priority': pandera.Column(pandera.String, checks=[
            pandera.Check.isin(list(WORK_ORDER_PRIORITY.__args__), ignore_na=False)], required=True),
        'Status': pandera.Column(pandera.String, checks=[
            pandera.Check.isin(list(WORK_ORDER_STATUS.__args__), ignore_na=False)], required=True),
        'CreatedDate': pandera.Column(pandera.DateTime, required=True, nullable=False),
        'LastModifiedDate': pandera.Column(pandera.DateTime, required=True, nullable=True),
        'WorkStartedDate': pandera.Column(pandera.DateTime, required=True, nullable=True),
        'WorkCompletedDate': pandera.Column(pandera.DateTime, required=True, nullable=True),
        'ClosedDate': pandera.Column(pandera.DateTime, required=True, nullable=True),
        'RawPriority': pandera.Column(pandera.String, required=True, nullable=False),
        'RawWorkOrderCategory': pandera.Column(pandera.String, required=True, nullable=False),
        'SubType': pandera.Column(pandera.String, required=False, nullable=True),
        'Vendor': pandera.Column(pandera.String, required=False, nullable=True),
        'VendorId': pandera.Column(pandera.String, required=False, nullable=True),
        'EquipmentClass': pandera.Column(pandera.String, required=False, nullable=True),
        'RawEquipmentClass': pandera.Column(pandera.String, required=False, nullable=True),
        'EquipmentLabel': pandera.Column(pandera.String, required=False, nullable=True),
        'RawEquipmentId': pandera.Column(pandera.String, required=False, nullable=True),
        'TenantId': pandera.Column(pandera.String, required=False, nullable=True),
        'TenantName': pandera.Column(pandera.String, required=False, nullable=True),
        'NotToExceedCost': pandera.Column(pandera.Float, required=False, nullable=True),
        'TotalCost': pandera.Column(pandera.Float, required=False, nullable=True),
        'BillableCost': pandera.Column(pandera.Float, required=False, nullable=True),
        'NonBillableCost': pandera.Column(pandera.Float, required=False, nullable=True),
        'Location': pandera.Column(pandera.String, required=False, nullable=True),
        'RawLocation': pandera.Column(pandera.String, required=False, nullable=True),
        'ScheduledStartDate': pandera.Column(pandera.DateTime, required=False, nullable=True),
        'ScheduledCompletionDate': pandera.Column(pandera.DateTime, required=False, nullable=True),
    },
    coerce=True,
    strict=False,
    name=None
)
